//  ===================================================================
//
//  Program:  commonlib
//  Module:   ludcmp.cpp
//  Language: C++
//  Date:     June 2004
//  Version:  1.0.0
//
//  Copyright (c) 2004
//
//  Marcelo Siqueira
//
//  Graduate Student
//  Department of Computer and Information Science
//  Moore School Building
//  200 South 33rd Street, Suite #62
//  Philadelphia, PA 19104-6389
//  Phone No.: (215) 898-5879
//
//  marcelos@seas.upenn.edu
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#include "ludcmp.h"

#include "macros.h"

#include <cmath>
#include <sstream>


namespace common {


  // -----------------------------------------------------------------
  // Method: tLUdcmp()
  // 
  // Default constructor
  //
  // -----------------------------------------------------------------
  tLUdcmp::tLUdcmp(
                   double** mat,
                   unsigned n
                  )
  {
    m_nele = n;
    m_matA = allocate(m_nele);

    unsigned int i;
    for (i=0; i<m_nele; ++i) {
      unsigned int j;
      for (j=0; j<m_nele; ++j) {
        m_matA[i][j] = mat[i][j];       
      }
    }

    m_perm.resize(m_nele);
    m_sign = 1;

    decomp();
  }


  // -----------------------------------------------------------------
  // Method: tLUdcmp()
  // 
  // Copy constructor
  //
  // -----------------------------------------------------------------
  tLUdcmp::tLUdcmp(const tLUdcmp& lu) : m_nele(lu.m_nele)
  {
    m_matA = allocate(m_nele);

    m_perm.resize(m_nele);
    m_sign = lu.m_sign;

    unsigned int i;
    for (i=0; i<m_nele; ++i) {
      m_perm[i] = lu.m_perm[i];

      unsigned int j;
      for (j=0; j<m_nele; ++j) {
        m_matA[i][j] = lu.m_matA[i][j];        
      }
    }
  }


  // -----------------------------------------------------------------
  // Method: ~tLUdcmp()
  // 
  // Destructor
  //
  // -----------------------------------------------------------------
  tLUdcmp::~tLUdcmp()
  {
    deallocate(m_matA,m_nele);
    m_matA = 0;
  }


  // -----------------------------------------------------------------
  // Method: allocate()
  // 
  // Allocates memory for a square matrix.
  //
  // -----------------------------------------------------------------
  double** 
  tLUdcmp::allocate(unsigned n)
  {
    double** mat = (double **) new double*[n];

    unsigned int i;
    for (i=0; i<n; ++i) {
      mat[i] = (double *) new double[n];
    }

    return mat;
  }


  // -----------------------------------------------------------------
  // Method: deallocate()
  // 
  // Releases memory held by a square matrix.
  //
  // -----------------------------------------------------------------
  void
  tLUdcmp::deallocate(
                      double** mat,
                      unsigned n
                     )
  {
    unsigned int i;
    for (i=0; i<n; ++i) {
      delete[] mat[i];
    }

    delete mat;
  }


  // -----------------------------------------------------------------
  // Method: decomp()
  // 
  // Performs LU decomposition in the matrix "m_matA"
  //
  // -----------------------------------------------------------------
  void
  tLUdcmp::decomp() throw (tExceptionObject)
  {
    const double TINY = 1.0e-20;
    int i, imax=0, j, k;
    double big, dum, sum, temp;

    int n = (int) m_nele;
    std::vector<double> vv(n);
    m_sign = 1;
    for (i=0; i<n; i++) {
      big=0.0;
      for (j=0; j<n; j++) {
        if ((temp=fabs(m_matA[i][j])) > big) {
          big=temp;
        }
      }

      if (big == 0.0) {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Cannot perform LU decomposition: matrix is singular";
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }

      vv[i] = 1.0 / big;
    }

    for (j=0; j<n; j++) {
      for (i=0; i<j; i++) {
        sum = m_matA[i][j];
        for (k=0; k<i; k++) {
          sum -= m_matA[i][k] * m_matA[k][j];
        }

        m_matA[i][j] = sum;
      }

      big = 0.0;

      for (i=j; i<n; i++) {
        sum = m_matA[i][j];
        for (k=0; k<j; k++) {
          sum -= m_matA[i][k] * m_matA[k][j];
        }

        m_matA[i][j] = sum;

        if ((dum = vv[i] * fabs(sum)) >= big) {
          big = dum;
          imax = i;
	}
      }

      if (j != imax) {
        for (k=0; k<n; k++) {
          dum = m_matA[imax][k];
          m_matA[imax][k] = m_matA[j][k];
          m_matA[j][k] = dum;
	}

        m_sign = -m_sign;
        vv[imax] = vv[j];
      }

      m_perm[j] = imax;
      if (m_matA[j][j] == 0.0) {
        m_matA[j][j] = TINY;
      }

      if (j != n-1) {
        dum = 1.0 / (m_matA[j][j]);
        for (i=j+1; i<n; i++) {
          m_matA[i][j] *= dum;
        }
      }
    }
  }


  // -----------------------------------------------------------------
  // Method: solve()
  // 
  // Solves a linear system AX = B, where A is the matrix "m_matA" and
  // B is a given vector. The solution vector is returned in B.
  //
  // -----------------------------------------------------------------
  void
  tLUdcmp::solve(
                 double* b,
                 unsigned n
                )
    const
  {
    if (n != m_nele) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Solution vector does not have expected dimension";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    int i, ii = 0, ip, j;
    double sum;
    for (i=0; i<(int)n; i++) {
      ip = m_perm[i];
      sum = b[ip];
      b[ip] = b[i];
      if (ii != 0) {
        for (j=ii-1; j<i; j++) {
          sum -= m_matA[i][j] * b[j];
        }
      }
      else if (sum != 0.0) {
        ii = i + 1;
      }

      b[i] = sum;
    }

    for (i=n-1; i>=0; i--) {
      sum = b[i];
      for (j=i+1; j<(int)n; j++) {
        sum -= m_matA[i][j] * b[j];
      }

      b[i] = sum / m_matA[i][i];
    }
  }


}        // end of namespace common
