/** 
 * @class dcel::tComplex
 *  
 * @brief A  class implementing a two-dimensional cell  complex in the
 * DCEL data structure.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#ifndef COMPLEX_H
#define COMPLEX_H

#include "exceptionobj.h"
#include "smartptr.h"
#include "dataobj.h"

#include "vertex.h"
#include "halfedge.h"
#include "edge.h"
#include "face.h"

#include "cpattrib.h"
#include "vtattrib.h"
#include "heattrib.h"
#include "egattrib.h"
#include "fcattrib.h"

#include <list>
#include <map>


namespace dcel {


  /**
   * Class  tComplex:  implements a  cell  complex  in  the DCEL  data
   * structure.
   *
   */ 
  class tComplex : public tDataObject {
  public:
    /**
     * Defines Self as this class.
     */    
    typedef tComplex              Self;


    /**
     * Defines Pointer as smart pointer to this class.
     */    
    typedef tSmartPtr<Self>       Pointer;


    /**
     * Defines Pointer as a const smart pointer to this class.
     */    
    typedef tSmartPtr<const Self> ConstPointer;


    /** 
     * Defines fcPointer as smart pointer to tFace.
     */    
    typedef tFace::Pointer        fcPointer;


    /** 
     * Defines egPointer as smart pointer to tEdge.
     */    
    typedef tEdge::Pointer        egPointer;


    /** 
     * Defines hePointer as smart pointer to tHalfEdge.
     */    
    typedef tHalfEdge::Pointer    hePointer;


    /** 
     * Defines vtPointer as smart pointer to tVertex.
     */    
    typedef tVertex::Pointer      vtPointer;


    /** 
     * Defines vtIterator as a iterator of a list of vtPointer.
     */    
    typedef std::list<vtPointer>::const_iterator vtIterator;


    /** 
     * Defines egIterator as a iterator of a list of egPointer.
     */    
    typedef std::list<egPointer>::const_iterator egIterator;


    /** 
     * Defines fcIterator as a iterator of a list of fcPointer.
     */    
    typedef std::list<fcPointer>::const_iterator fcIterator;


    /**
     * Defines atPointer as a smart pointer to a tCPAttrib object.
     */
    typedef tSmartPtr<tCPAttrib>                 atPointer;


    /**
     * Defines vaPointer as a smart pointer to a tVTAttrib object.
     */
    typedef tSmartPtr<tVTAttrib>                 vaPointer;


    /**
     * Defines eaPointer as a smart pointer to a tEGAttrib object.
     */
    typedef tSmartPtr<tEGAttrib>                 eaPointer;


    /**
     * Defines haPointer as a smart pointer to a tHEAttrib object.
     */
    typedef tSmartPtr<tHEAttrib>                 haPointer;


    /**
     * Defines faPointer as a smart pointer to a tFCAttrib object.
     */
    typedef tSmartPtr<tFCAttrib>                 faPointer;


    /** 
     * Defines VMAP as a hash table of vertex IDs and vertex pointers.
     */
    typedef std::map<unsigned,vtPointer> VMAP;


    /** 
     * Defines HMAP  as a  hash table of  half-edge IDs  and half-edge
     * pointers.
     */
    typedef std::map<std::pair<unsigned,unsigned>,hePointer> HMAP;
 

    /**
     * Method for creating an instance of this object.
     *
     * @param a: a pointer to a set of cell complex attributes.
     *
     */
    static Pointer NEW(const atPointer& a)
    {
      tComplex *dumbPtr = new tComplex;
      Pointer smartPtr(dumbPtr);
      smartPtr->set_attrib(a);

      return smartPtr;
    }


    /** 
     * Virtual destructor.
     */
    virtual ~tComplex() {
    }


    /** 
     * Builds the cell complex from a set of vertices and faces.
     *
     */
    void build(unsigned, double*, unsigned, unsigned*, unsigned)
      throw (tExceptionObject);


    /** 
     * Check if the cell complex has a boundary. 
     *
     */
    bool has_boundary() const;


    /** 
     * Returns the number of vertices in this cell complex.
     *
     * @return The number of vertices in this cell complex.
     *
     */
    unsigned get_num_verts() const
      { return m_lverts.size(); }


    /** 
     * Returns the number of edges in this cell complex.
     *
     * @return The number of edges in this cell complex.
     *
     */
    unsigned get_num_edges() const
      { return m_ledges.size(); }


    /**
     * Returns the number of faces in this cell complex.
     *
     * @return The number of faces in this cell complex.
     *
     */
    unsigned get_num_faces() const
      { return m_lfaces.size(); }


    /**
     * Returns a pointer to the set of attributes of this complex.
     *
     * @return A pointer to a set of cell complex attributes.
     *
     */
    atPointer get_attrib() const
      { return m_at; }


    /**
     * Assigns a given pointer to the pointer to the set of attributes
     * of this face.
     *
     * @param at: a pointer to a set of complex attributes.
     *
     */
    void set_attrib(const atPointer& at)
      { m_at = at; }


    /**
     * Returns     the    initial     position     of    the     @link
     * dcel::tComplex::vtIterator vertex iterator @endlink.
     *
     * @return A vertex  list iterator set to the  initial position of
     * the list.
     *
     */
    vtIterator vt_begin() const { return m_lverts.begin();  }


    /**
     * Returns     the    initial     position     of    the     @link
     * dcel::tComplex::egIterator edge iterator @endlink.
     *
     * @return An  edge list iterator  set to the initial  position of
     * the list.
     *
     */
    egIterator eg_begin() const { return m_ledges.begin();  }


    /**
     * Returns     the    initial     position     of    the     @link
     * dcel::tComplex::fcIterator face iterator @endlink.
     *
     * @return A  facet list iterator  set to the initial  position of
     * the list.
     */
    fcIterator fc_begin() const { return m_lfaces.begin();  }

 
    /**
     * Returns     the     last      position     of     the     @link
     * dcel::tComplex::vtIterator vertex iterator @endlink.
     *
     * @return A vertex list iterator set to the final position of the
     * list.
     *
     */
    vtIterator vt_end() const { return m_lverts.end();  }


    /**
     * Returns     the     last      position     of     the     @link
     * dcel::tComplex::egIterator edge iterator @endlink.
     *
     * @return An edge list iterator  set to the final position of the
     * list.
     *
     */
    egIterator eg_end() const { return m_ledges.end();  }


    /**
     * Returns     the     last      position     of     the     @link
     * dcel::tComplex::fcIterator face iterator @endlink.
     *
     * @return A face  list iterator set to the  final position of the
     * list.
     *
     */
    fcIterator fc_end() const { return m_lfaces.end();  }


    /** 
     * Verifies whether the cell complex is a simplicial complex.
     *
     */
    bool is_simplicial() const;


    /** 
     * Removes all smart pointer  references in the structure, so that
     * all  circular  dependencies  are  removes  and  memory  can  be
     * released correctly.
     *
     */
    void cleanup();


  private:
    /**
     * Default constructor.
     */
    tComplex() {}


    /** 
     * Creates the vertices of a cell complex.
     *
     */
    void create_vertices(unsigned, double*, VMAP&);


    /** 
     * Creates the faces of a cell complex.
     *
     */
    void create_faces(unsigned, unsigned*, unsigned, HMAP&, 
      const VMAP&) throw (tExceptionObject);


    /** 
     * Creates the edges of a cell complex.
     *
     */
    void create_edges(HMAP&);


    /** 
     * Performs a topological consistency check in a cell complex.
     *
     */
    bool is_consistent() const;


    /**
    * List of all vertices in this complex.
    */
    std::list<vtPointer> m_lverts;


    /**
    * List of all edges in this complex.
    */
    std::list<egPointer> m_ledges;


    /**
    * List of all faces in this complex.
    */
    std::list<fcPointer> m_lfaces;


    /**
     * A pointer to the set of attributes of this complex.
     */
    atPointer m_at;

  };


}


#endif   // COMPLEX_H
