/** 
 * @class dcel::tHalfEdge
 *  
 * @brief A class implementing a halfedge of the DCEL data structure.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#ifndef HALFEDGE_H_
#define HALFEDGE_H_

#include "smartptr.h"
#include "dataobj.h"
#include "heattrib.h"
#include "vertex.h"


namespace dcel {


  /**
   * Forward definition of classes.
   */
  class tComplex;
  class tEdge;
  class tFace;


  /**
   * Class  tHalfEdge:   implements  a  halfedge  of   the  DCEL  data
   * structure.
   *
   */ 
  class tHalfEdge : public tDataObject {
  public:
    /**
     * Defines Self as this class.
     */   
    typedef tHalfEdge             Self;


    /**
     * Defines Pointer as smart pointer to this class.
     */ 
    typedef tSmartPtr<Self>       Pointer;


    /**
     * Defines ConstPointer as a const smart pointer to this class.
     */  
    typedef tSmartPtr<const Self> ConstPointer;


    /** 
     * Defines vtPointer as smart pointer to tVertex.
     */    
    typedef tVertex::Pointer      vtPointer;


    /** 
     * Defines egPointer as smart pointer to tEdge.
     */    
    typedef tSmartPtr<tEdge>      egPointer;


    /** 
     * Defines fcPointer as smart pointer to tFace.
     */    
    typedef tSmartPtr<tFace>      fcPointer;


    /**
     * Defines atPointer as a smart pointer to a tHEAttrib object.
     */
    typedef tSmartPtr<tHEAttrib>  atPointer;


    /** 
     * Method for creating an instance of this object.
     *
     * @param v: a pointer to a vertex.
     * @param e: a pointer to an edge.
     * @param f: a pointer to a facet.
     * @param n: a pointer to a half-edge.
     * @param p: a pointer to a half-edge.
     * @param a: a pointer to a set of half-edge attributes.
     *
     * @return  A pointer to  a newly  created half-edge.
     *
     */
    static Pointer NEW(
                       const vtPointer& v,
                       const egPointer& e,
                       const fcPointer& f,
                       const   Pointer& n,
                       const   Pointer& p,
                       const atPointer& a
                      )
    {
      tHalfEdge *dumbPtr = new tHalfEdge;
      Pointer smartPtr(dumbPtr);

      smartPtr->set_vert(v);
      smartPtr->set_edge(e);
      smartPtr->set_face(f);
      smartPtr->set_next(n);
      smartPtr->set_prev(p);
      smartPtr->set_attrib(a);

      return smartPtr;
    }


    /** 
     * Virtual destructor.
     */
    ~tHalfEdge() {}


    /** 
     * Returns a pointer to the origin vertex of this half-edge.
     *
     * @return A pointer to a vertex.
     *
     */
    vtPointer get_vert() const
      { return m_vert; }

    
    /** 
     * Returns a pointer to the edge this half-edge belongs to.
     *
     * @return A pointer to an edge.
     *
     */
    egPointer get_edge() const
      { return m_edge; }

    
    /** 
     * Returns a pointer to the face this half-edge belongs to.
     *
     * @return A pointer to a face.
     *
     */
    fcPointer get_face() const
      { return m_face; }


    /** 
     * Returns a pointer to the  half-edge following this one in their
     * face cycle.
     *
     * @return A pointer to a half-edge.
     *
     */
    Pointer next() const
      { return m_next; }


    /** 
     * Returns a pointer to the  half-edge preceding this one in their
     * face cycle.
     *
     * @return A pointer to a half-edge.
     *
     */
    Pointer prev() const
      { return m_prev; }


    /**
     * Returns a pointer to the set of attributes of this half-edge.
     *
     * @return A pointer to a set of half-edge attributes.
     *
     */
    atPointer get_attrib() const
      { return m_at; }


    /**
     * Assigns a given pointer to the pointer to the set of attributes
     * of this half-edge.
     *
     * @param at: a pointer to a set of half-edge attributes.
     *
     */
    void set_attrib(const atPointer& at)
      { m_at = at; }


    /**
     * Makes tComplex a friend class.
     *
     */
    friend class tComplex;


  private:
    /** 
     * Default constructor.            
     */
    tHalfEdge() {}


    /** 
     * Assigns a given pointer to  the pointer to the origin vertex of
     * this half-edge.
     *
     * @param v: a pointer to a vertex.
     */
    void set_vert(const vtPointer& v)
      { m_vert = v; }


    /** 
     * Assigns  a  given pointer  to  the  pointer  to the  edge  this
     * half-edge belongs to.
     *
     * @param e: a pointer to an edge.
     */
    void set_edge(const egPointer& e)
      { m_edge = e; }
    

    /** 
     * Assigns  a  given pointer  to  the  pointer  to the  face  this
     * half-edge belongs to.
     *
     * @param f: a pointer to a facet.
     *
     */
    void set_face(const fcPointer& f)
      { m_face = f; }


    /** 
     * Assigns a given pointer to the pointer to the successor of this
     * half-edge in its face cycle.
     *
     * @param he: a pointer to a half-edge.
     *
     */
    void set_next(const Pointer& he)
      { m_next = he; }
    

    /** 
      * Assigns a given  pointer to the pointer to  the predecessor of
      * this half-edge in its face cycle.
      *
      * @param he: a pointer to a half-edge.
      *
      */
    void set_prev(const Pointer& he)
      { m_prev = he; }


    /**
     *  A pointer to the origin vertex of this half-edge.
     */
    vtPointer m_vert;


    /**
     *  A pointer to the edge this half-edge belongs to.
     */    
    egPointer m_edge;


    /**
     * A pointer to the face this half-edge belongs to.
     */    
    fcPointer m_face;


    /**
     * A pointer to  the next half-edge in the cycle  of the face this
     * half-edge belongs to.
     *
     */    
    Pointer m_next;


    /**
     *  A pointer to  the previous half-edge in the  cycle of the face
     *  this half-edge belongs to.
     *
     */    
    Pointer m_prev;


    /**
     * A pointer to the set of attributes of this half-edge.
     */
    atPointer m_at;

  };


}
#endif   // HALFEDGE_H_
