/** 
 * @class dcel::tSmartPtr
 *  
 * @brief A smart pointer class
 *
 *
 * @details This class is an implementation of smart pointers according
 *          to Scott Meyers' book "More Effective C++", Addison-Wesley,
 *          1996, pp. 159-182. For reference counting purposes, it must
 *          be used only with classes that inherits from tDataObject.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#ifndef SMARTPTR_H_
#define SMARTPTR_H_


/**
 * Class tSmartPtr: implements a smart pointer.
 *
 */ 
template <class T>
class tSmartPtr {
public:
  /**
   * Constructs  a  smart   pointer  initialized  with  the  specified
   * pointer.
   *
   */
  explicit tSmartPtr<T>(T* ptr = 0);

    
  /**
   * Constructs  a  new smart  pointer  copying  the  contents of  the
   * specified one.
   *
   */
  tSmartPtr<T>(const tSmartPtr<T>&);


  /**
   * Destructor.
   */
  ~tSmartPtr<T>();


  /**
   * Assignment operator.
   *
   */
  tSmartPtr<T>& operator=(const tSmartPtr<T>&); 


  /**
   * Deferencing operator.
   *
   */
  T& operator*() const;


  /**
   * Deferencing operator.
   *
   */
  T* operator->() const;


  /**
   * Overload operator to test equality.
   *
   */
  bool operator==(const tSmartPtr<T>&) const;


  /**
   * Overload operator to test equality.
   *
   */
  bool operator!=(const tSmartPtr<T>&) const;


  /**
   * Overload operator less than.
   *
   */
  bool operator<(const tSmartPtr<T>&) const;


  /**
   * Test nulliness of the pointee.
   *
   */
  bool operator!() const;

    
  /**
   * Test nulliness of the pointee.
   *
   * @return True if the pointer is not @c null and false otherwise.
   */
  bool is_not_null() const;

    
  /**
   * Test nulliness of the pointee.
   *
   */
  bool is_null() const;


  /**
   * Returns the native C pointer.
   *
   */
  T* get_pointee() const;


  /**
   * Template function for implicit conversion operations.
   *
   */
  template<class newT> operator tSmartPtr<newT>() {
    return tSmartPtr<newT>(m_pointee);
  }


private:
  /**
   * Calls  the method that  decrements the  reference counter  of the
   * object pointed by this smart pointer.
   *
   */
  void release();


  /**
   * Actual pointer to an object of type T.
   */
  T* m_pointee;
};


/**
 * Constructs a smart pointer initialized with the specified pointer.
 *
 * @param ptr: a native C pointer to an object of type T.
 *
 */
template <class T>
tSmartPtr<T>::tSmartPtr(T* ptr)
{
  m_pointee = ptr;
  if (m_pointee != 0) {
    m_pointee->add_ref(); 
  }
}


/**
 * Constructs  a  new  smart  pointer  copying  the  contents  of  the
 * specified one.
 *
 * @param ptr: a smart pointer to an object of type T.
 *
 */
template <class T>
tSmartPtr<T>::tSmartPtr(const tSmartPtr<T>& ptr)
{
  m_pointee = ptr.m_pointee;
  if (m_pointee != 0) {
    m_pointee->add_ref();
  }
}


/**
 * Destructor.
 */
template <class T>
tSmartPtr<T>::~tSmartPtr<T>() 
{
  release();
}


/**
 * Assignment operator.
 *
 * @param ptr: a smart pointer to an object of type T.
 *
 * @return This instance.
 *
 */
template <class T>
tSmartPtr<T>&
tSmartPtr<T>::operator=(const tSmartPtr<T>& ptr)
{
  release();

  m_pointee = ptr.m_pointee;

  if (m_pointee != 0) {
    m_pointee->add_ref();
  }

  return *this;
}


/**
 * Deferencing operator.
 *
 * @return The actual pointed object.
 *
 */
template <class T>
T&
tSmartPtr<T>::operator*() const
{
  return *(m_pointee);
}


/**
 * Deferencing operator.
 *
 * @return The actual native C pointer.
 *
 */
template <class T>
T*
tSmartPtr<T>::operator->() const
{
  return m_pointee;
}


/**
 * Overload operator to test equality.
 *
 * @return True if  both smart pointers point to  the same object, and
 * false otherwise.
 *
 */
template <class T>
bool
tSmartPtr<T>::operator==(const tSmartPtr<T>& ptr) const
{
  return m_pointee == ptr.m_pointee;
}


/**
 * Overload operator to test equality.
 *
 * @return False if both smart  pointers point to the same object, and
 * true otherwise.
 *
 */
template <class T>
bool
tSmartPtr<T>::operator!=(const tSmartPtr<T>& ptr) const
{
  return m_pointee != ptr.m_pointee;
}


/**
 * Overload operator less than.
 *
 * @return True if the address of the pointee of this smart pointer is
 * less than  the address of the  pointee of the  given smart pointer;
 * otherwise, return false.
 *
 */
template <class T>
bool
tSmartPtr<T>::operator<(const tSmartPtr<T>& ptr) const
{
  return m_pointee < ptr.m_pointee;
}


/**
 * Test nulliness of the pointee.
 *
 * @return True if the pointer is @c null and false otherwise.
 *
 */
template <class T>
bool
tSmartPtr<T>::operator!() const
{
  return (m_pointee == 0);
}


/**
 * Test nulliness of the pointee.
 *
 * @return True if the pointer is not @c null and false otherwise.
 *
 */
template <class T>
bool
tSmartPtr<T>::is_not_null() const
{
  return (m_pointee != 0);
}


/**
 * Test nulliness of the pointee.
 *
 * @return True if the pointer is @c null and false otherwise.
 *
 */
template <class T>
bool
tSmartPtr<T>::is_null() const
{
  return (m_pointee == 0);
}


/**
 * Returns the native C pointer.
 *
 * @return The native C pointer encapsulated by this smart pointer.
 *
 */
template <class T>
T*
tSmartPtr<T>::get_pointee() const
{
  return m_pointee;
}
  

/**
 * Calls  the method  that  decrements the  reference  counter of  the
 * object pointed by this smart pointer.
 *
 */
template <class T>
void
tSmartPtr<T>::release()
{
  T* temp = m_pointee;

  if (temp) temp->release();

  m_pointee = 0;
}


#endif  // SMARTPTR_H_
