//  ===================================================================
//
//  Program:  OFFLIB
//  Module:   writer.cpp
//  Language: C++
//  Date:     February 2008
//  Version:  1.0.0
//
//  Copyright (c) 2008
//
//  Marcelo Siqueira
//
//  Departamento de Computao e Estatstica
//  Universidade Federal de Mato Grosso do Sul
//  Av. Costa e Silva, S/N, Campus Universitria
//  Campo Grande, MS, Brasil, 79090-700
//  Phone No.: +55 67 3345 7458
//             +55 67 3345 7515
//
//  marcelo@dct.ufms.br
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ==================================================================

#include <iostream>
#include <iomanip>
#include <sstream>

#include "writer.h"


namespace offlib {


  // -----------------------------------------------------------------
  // Method: tWriter()
  // 
  // Default constructor
  //
  // -----------------------------------------------------------------
  tWriter::tWriter(const std::string& fn) throw (tExceptionObject)
  {
    if (fn.empty()) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Invalid implicit object name";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }
    
    m_fname = fn + std::string(".off");
  }


  // -----------------------------------------------------------------
  // Method: tWriter()
  // 
  // Copy constructor
  //
  // -----------------------------------------------------------------
  tWriter::tWriter(const tWriter& w) : m_fname (w.m_fname)
  {}


  // -----------------------------------------------------------------
  // Method: ~tWriter()
  // 
  // Destructor
  //
  // -----------------------------------------------------------------
  tWriter::~tWriter()
  {}


  // -----------------------------------------------------------------
  // Method: write()
  // 
  // Calls the methods for writing the vertex, edge, and face files.
  //
  // -----------------------------------------------------------------
  void
  tWriter::write(
                 unsigned nv,
                 const double* vset,
                 unsigned nf,
                 const unsigned* fset
		)
    throw (tExceptionObject)
  {
    m_fs.open(m_fname.c_str(), std::ios::out | std::ios::binary);
    if (!m_fs) {
      std::stringstream ss (std::stringstream::in | 
			    std::stringstream::out);
      ss << "Cannot open file " << m_fname;
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    write_header(nv,nf);
    write_verts(nv,vset);
    write_faces(nf,fset);

    // close file
    m_fs.close();
  }


  // -----------------------------------------------------------------
  // Method: write_qm()
  // 
  // Calls methods for writing the vertex, edge, and face files.
  //
  // -----------------------------------------------------------------
  void
  tWriter::write_qm(
                     unsigned nv,
                     const double* vset,
                     unsigned nf,
                     const unsigned* fset
		    )
    throw (tExceptionObject)
  {
    m_fs.open(m_fname.c_str(), std::ios::out | std::ios::binary);
    if (!m_fs) {
      std::stringstream ss (std::stringstream::in | 
			    std::stringstream::out);
      ss << "Cannot open file " << m_fname;
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    write_header(nv,nf);
    write_verts(nv,vset);

    write_faces_qm(nf,fset);

    // close file
    m_fs.close();
  }


  // -----------------------------------------------------------------
  // Method: write_header()
  // 
  // Writes the number of vertices, edges, and faces of the triangular
  // mesh.
  //
  // -----------------------------------------------------------------
  void
  tWriter::write_header(
                        unsigned nv,
                        unsigned nf
                       )
  {
    //
    // Write first line
    //
    m_fs << "OFF" << std::endl;

    //
    // Write the number of vertices, faces, and edges
    //
    m_fs << nv
         << " "
         << nf
         << " 0"
         << std::endl;
  }


  // -----------------------------------------------------------------
  // Method: write_verts()
  // 
  // Writes  the vertex  information of  a triangular  mesh to  an OFF
  // file.
  //
  // -----------------------------------------------------------------
  void
  tWriter::write_verts(
                       unsigned nv,
                       const double* vset
                      ) 
    throw (tExceptionObject)
  {
    m_fs << std::fixed << std::setprecision(18) ;

    unsigned i;
    for (i=0; i<nv; i++) {
      const unsigned j = 3 * i;

      m_fs << vset[j]
           << " "
           << vset[j + 1] 
           << " "
           << vset[j + 2]
           << std::endl;
    }
  }


  // -----------------------------------------------------------------
  // Method: write_faces()
  // 
  // Writes the face information of a triangular mesh to an OFF file.
  //
  // -----------------------------------------------------------------
  void
  tWriter::write_faces(
                       unsigned nf,
                       const unsigned* fset
                      ) 
    throw (tExceptionObject)
  {
    unsigned i;
    for (i=0; i<nf; i++) {
      const unsigned j = 3 * i;

      m_fs << "3 "
           << fset[j]
           << " "
           << fset[j + 1]
           << " "
           << fset[j + 2]
           << std::endl;
    }
  }


  // -----------------------------------------------------------------
  // Method: write_faces_qm()
  // 
  // Writes the  face information  of a quadrilateral  mesh to  an OFF
  // file.
  //
  // -----------------------------------------------------------------
  void
  tWriter::write_faces_qm(
                          unsigned nf,
                          const unsigned* fset
                         ) 
    throw (tExceptionObject)
  {
    unsigned i;
    for (i=0; i<nf; i++) {
      const unsigned j = 4 * i;

      m_fs << "4 "
           << fset[j]
           << " "
           << fset[j + 1]
           << " "
           << fset[j + 2]
           << " "
           << fset[j + 3]
           << " "
           << std::endl;
    }
  }


}    // end namespace offlib
