/** 
 * @class pps::tBezier
 *  
 * @brief A  class implementing a rectangular Bezier  surface patch in
 * 3D.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date May 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#include "bezier.h"

#include <sstream>
#include <cmath>
#include <cassert>
#include <stdlib.h>

#include "mathmacros.h"
#include "ludcmp.h"


namespace pps {


  using common::tLUdcmp;


  /**
   * Creates  a  rectangular  Bezier   patch  from  a  set  of  sample
   * points. The  function uses a  least squares procedure  to compute
   * the Bezier coefficients.
   *
   * @param pp: an array with the 2D coordinates of a set of parameter
   * points.
   * @param sp: an array with the 3D coordinates of a set of parameter
   * 3D points, which  are the image of the  paramenter points under a
   * possibly unknown function.
   * @param np: the number of parameter points.
   * @param n: the first degree of this rectangular Bezier patch.
   * @param m: the second degree of this rectangular Bezier patch.
   * @param rx: the first coordinate  of the first point of the affine
   * frame w.r.t. which this surface patch is defined.
   * @param sx: the second coordinate of the first point of the affine
   * frame w.r.t. which this surface patch is defined.
   * @param rx: the first coordinate of the second point of the affine
   * frame w.r.t. which this surface patch is defined.
   * @param  sx: the  second coordinate  of  the second  point of  the
   * affine frame w.r.t. which this surface patch is defined.
   */
  void
  tBezier::build(
                 double* pp,
                 double* sp,
                 unsigned np,
                 unsigned m,
                 unsigned n,
                 double rx,
                 double sx,
                 double ry,
                 double sy
		)
    throw (tExceptionObject) 
  {
    /**
     * Check consistency of the input parameters.
     */
    m_m = m;
    m_n = n;

    if ( (m_m < 1 ) || ( m_n < 1 ) ) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "The patch bi-degree must be at least (1,1)"; 
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    m_rx = rx;
    m_sx = sx;

    if ( ( m_rx >= m_sx ) ) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "The given affine frame is inconsistent";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    m_ry = ry;
    m_sy = sy;

    if ( ( m_ry >= m_sy ) ) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "The given affine frame is inconsistent";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    if ( (2 * (m_n + 1) * (m_m + 1) ) > np ) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "The number of sample points is too small";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    /**
     * Allocate memory for the control points.
     */
    unsigned ncp = (m_m + 1) * (m_n + 1);

    m_b = new double*[ ( m_m + 1 ) * ( m_n + 1 ) ];

    unsigned i;
    for (i=0; i<ncp; i++) {
      m_b[i] = new double[3];
    }

    /**
     * Compute a (np x ncp)  matrix "A" with all Bernstein polynomials
     * of order "m_m + m_n" at the parameter points given in the input
     * vector "pp".
     */

    double** A = new double*[np];

    for (i=0; i<np; i++) {
      A[i] = new double[ncp];
    }

    comp_bpoly_matrix( A , pp , np );

    /**
     * Compute the  product "At x A",  where "At" is  the transpose of
     * "A".
     */
    double** AtA = new double*[ncp];

    for (i=0; i<ncp; i++) {
      AtA[i] = new double[ncp];
    }

    comp_matrix_ata( A , np , ncp , AtA );

    /**
     * Compute the  product "At x B",  where "At" is  the transpose of
     * "A".
     */
    double** AtB = new double*[ncp];

    for (i=0; i<ncp; i++) {
      AtB[i] = new double[3];
    }

    comp_matrix_atb( A , sp , np , ncp , AtB );

    /**
     * Solve the linear system "AtA Y = AtB[i]", where "AtB[i]" is the
     * i-th column  of "AtB",  for i=0,1,2. This  is done by  using LU
     * decomposition.
     */
    
    /**
     * Compute the LU decomposition of AtA.
     */
    tLUdcmp LU( AtA , ncp );


    /**
     * Allocate  memory for the  solution vector,  "Y", of  the linear
     * system "AtA Y = AtB[i].
     *
     */

    double* Y = new double[ncp];

    for (i=0; i<3; i++) {
      unsigned j;
      for (j=0; j<ncp; j++) {
        Y[j] = AtB[j][i];
      }

      /**
       * Solve the linear system AtA Y = AtB[i].
       */
      try {
        LU.solve(Y,ncp); 
      }
      catch (const tExceptionObject& xpt) {
        treatException(xpt);
        exit(0);
      }

      /**
       * Store the solution in the i-th column of the matrix "m_b[j]".
       */
      for (j=0; j<ncp; j++) {
        m_b[j][i] = Y[j];
      }
    }

    /**
     * Release memory
     */    
    if (A != 0) {
      for (i=0; i<np; i++) {
        if (A[i] != 0) {
          delete[] A[i];
        }
      }

      delete A;
    }

    if (AtA != 0) {
      for (i=0; i<ncp; i++) {
        if (AtA[i] != 0) {
          delete[] AtA[i];
        }
      }

      delete AtA;
    }

    if (AtB != 0) {
      for (i=0; i<ncp; i++) {
        if (AtB[i] != 0) {
          delete[] AtB[i];
        }
      }

      delete AtB;
    }

    delete[] Y;
  }


  /**
   * Destructor.
   */
  tBezier::~tBezier()
  {
    if (m_b != 0) {
      unsigned ncp = (m_m + 1) * (m_n + 1);
      unsigned i;
      for (i=0; i<ncp; i++) {
        if (m_b[i] != 0) {
          delete[] m_b[i];
        }
      }

      delete[] m_b;
    }
  }


  /**
   * Given two indices,  "i" and "j", this method  returns the control
   * point "b_(i,j)" of this rectangular Bezier surface patch.
   *
   * @param i: first index of the control point.
   * @param j: second index of the control point.
   *
   * @return The  control point  "b_(i,j)" of this  rectangular Bezier
   * surface patch.
   */
  t3DVector
  tBezier::b(
             unsigned i,
             unsigned j
            ) 
    const throw (tExceptionObject)
  {
    if ( (i > m_m ) || ( j > m_n ) ) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Input parameter value is out of range";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    unsigned l = index(j,i);

    return t3DVector(m_b[l][0], m_b[l][1], m_b[l][2]);
  }


  /**
   * Given a parameter  value, this method computes the  point on this
   * Bezier patch  corresponding to the  given barycentric coordinates
   * of a parameter value.
   *
   * @param x: first rectangular coordinate of the parameter point.
   * @param y: second rectangular coordinate of the parameter point.
   *
   * @return  A point  on this  patch  that is  the image  of a  given
   * parameter point.
   *
   */
  t3DVector
  tBezier::point(
                 double x,
                 double y
                ) 
    const throw (tExceptionObject)
  {
    /**
     * Map the point to the affine frame [0,1].
     */
    double u = (x - m_rx) / (m_sx - m_rx);
    double v = (y - m_ry) / (m_sy - m_ry);

    if (ABS(u) <= 1e-15) {
      u = 0;
    }
    else if (ABS(1-u) <= 1e-15) {
      u = 1;
    }

    if (ABS(v) <= 1e-15) {
      v = 0;
    }
    else if (ABS(1-v) <= 1e-15) {
      v = 1;
    }

    /**
     * Compute all Bernstein polynomials of degree "m_m".
     */
    std::vector<double> bu(m_m+1);
    all_bernstein( m_m , u , bu );

    /**
     * Compute all Bernstein polynomials of degree "m_n".
     */
    std::vector<double> bv(m_n+1);
    all_bernstein( m_n , v , bv );

    /**
     * Compute the image of (u,v) on the surface patch.
     */
    t3DVector sum(0,0,0);
    unsigned j;
    for (j=0; j<=m_n; j++) {
      unsigned i;
      for (i=0; i<=m_m; i++) {
        double buv = bu[i] * bv[j];
        try {
          sum += ( buv * b( i , j ) );
        }
        catch (const tExceptionObject& xpt) {
          treatException(xpt);
          exit(0);
        }
      }
    }

    return sum;
  }


  /**
   * This  function computes  the  value of  one Bernstein  polynomial
   * (B_(n,i)) at a  given parameter value with respect  to the affine
   * frame [0,1].
   *
   * @param n: the degree of the Bernstein polynomial.
   * @param i: the index of the Bernstein polynomial of degree "n".
   * @param u: a parameter value.
   */
  double
  tBezier::bernstein(
                     unsigned n,
                     unsigned i,
                     double u
		    )
    const throw (tExceptionObject)
  {
    assert(i <= n);

    std::vector<double> temp(n+1);

    unsigned j;
    for (j=0; j<=n; j++) {
      temp[j] = 0;
    }

    temp[n-i] = 1;

    double u1 = 1 - u;

    unsigned k;
    for (k=1; k<=n; k++) {
      for (j=n; j>=k; j--) {
        temp[j] = (u1 * temp[j]) + (u * temp[j-1]);
      }
    }

    return temp[n];
  }


  /**
   * This function computes the  value of all Bernstein polynomials at
   * a given parameter value with respect to the affine frame [0,1].
   *
   * @param n: the degree of the Bernstein polynomial.
   * @param u: a parameter value.
   * @param b: an  array with the value of  all Berbnstein polynomials
   * at "u".
   */
  void
  tBezier::all_bernstein(
                         unsigned n,
                         double u,
                         std::vector<double>& b
		        )
    const throw (tExceptionObject)
  {
    b[0] = 1;
    double u1 = 1 - u;

    unsigned j;
    for (j=1; j<=n; j++) {
      double saved = 0;
      unsigned k;
      for (k=0; k<j; k++) {
        double temp = b[k];
        b[k] = saved + (u1 * temp);
        saved = u * temp;
      }

      b[j] = saved;
    }
  }


  /**
   * This  function computes  the  pairwise product  of all  Bernstein
   * polynomials  of order  "m_m" and  "m_n"  at a  set of  parameters
   * points.
   *
   * @param A: a (np  x ((m_m + 1) * (m_n +  1)) matrix such that each
   * element of  "A" is  a product of  two Bernstein polynomials  at a
   * given parameter point.
   * @param P: a vector containing the 2D coordinates of the parameter
   * points.
   * @param np: number of parameter points.
   *
   */
  void
  tBezier::comp_bpoly_matrix(
                             double**& A,
                             double* P,
                             unsigned np
		            )
    const
  {   
    /**
     * Compute the value of the Bernstein polynomials at the parameter
     * points.
     */

    unsigned i;
    for (i=0; i<np; i++) {
      double x = ( P[(i << 1)    ] - m_rx ) / ( m_sx - m_rx );
      double y = ( P[(i << 1) + 1] - m_ry ) / ( m_sy - m_ry );

      std::vector<double> bu(m_m+1);
      std::vector<double> bv(m_n+1);

      try {
        all_bernstein( m_m , x , bu );
        all_bernstein( m_n , y , bv );
      }
      catch (const tExceptionObject& xpt) {
        treatException(xpt);
        exit(0);
      }

      unsigned k;
      for (k=0; k<=m_n; ++k) {
        unsigned j;
        for (j=0; j<=m_m; ++j) {
          unsigned l = index( k , j );
          A[i][l] = bu[j] * bv[k];
        }
      }
    }      
  }


  /**
   * This function  computes the product  of the transpose of  a given
   * matrix with the matrix itself.
   *
   * @param A: a (n x p) matrix.
   * @param n: number of rows of "A".
   * @param p: number of columns of "A".
   * @param  AtA:  a (p  x  p)  matrix equal  to  the  product of  the
   * transpose of a given matrix with the matrix itself.
   *
   */
  void
  tBezier::comp_matrix_ata(
                           double** A,
                           unsigned n,
                           unsigned p,
                           double**& AtA
		          )
    const
  { 
    unsigned i;
    for (i=0; i<p; i++) {
      unsigned k;
      for (k=0; k<p; k++) {
        AtA[i][k] = 0;
        unsigned j;
        for (j=0; j<n; j++) {
          AtA[i][k] += A[j][i] * A[j][k];
        }
      }
    }
  }


  /**
   *  This function computes  the product of the transpose  of a given
   *  (n x p) matrix "A" and a (n  x 3) matrix B. The result is a (p x
   *  3) matrix.
   *
   * @param A: the first matrix.
   * @param B: the second matrix.
   * @param n: number of rows of the matrices.
   * @param p: number of columns of the first matrix.
   * @param AtB:  the matrix resulting from the  multiplication of the
   * transpose of the first matrix by the second matrix.
   *
   */
  void
  tBezier::comp_matrix_atb(
                           double** A,
                           double* B,
                           unsigned n,
                           unsigned p,
                           double**& AtB
		          )
    const
  {
    /**
     * Compute the  product (At x B),  where "At" is  the transpose of
     * "A".
     */
    unsigned i;
    for (i=0; i<p; i++) {
      AtB[i] = new double[3];
      unsigned k;
      for (k=0; k<3; k++) {
        AtB[i][k] = 0;
        unsigned j;
        for (j=0; j<n; j++) {
          AtB[i][k] += A[j][i] * B[(3 * j) + k];
        }
      }
    }
  }


}       // end of namespace pps
