/** 
 * @class pps::tBezier
 *  
 * @brief A  class implementing a rectangular Bezier  surface patch in
 * 3D.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date May 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#ifndef BEZIER_H
#define BEZIER_H

#include "smartptr.h"
#include "dataobj.h"
#include "macros.h"
#include "exceptionobj.h"

#include "3dvector.h"

#include <stdlib.h>
#include <vector>


namespace pps {


  using common::t3DVector;


  /**
   * Class tBezier: A class  implementing a rectangular Bezier surface
   * patch in 3D.
   *
   */

  class tBezier : public tDataObject {
  public:
    /**
     * Defines Self as this class.
     */
    typedef tBezier               Self;


    /**
     * Defines Pointer as a smart pointer to this class.
     */
    typedef tSmartPtr<Self>       Pointer;

    
    /**
     * Defines ConstPointer as a const smart pointer to this class.
     */
    typedef tSmartPtr<const Self> ConstPointer;


    /**
     * Method for creating an instance of this object.
     *
     * @param  pp:  an array  with  the 2D  coordinates  of  a set  of
     * parameter points.
     * @param  sp:  an array  with  the 3D  coordinates  of  a set  of
     * parameter  3D points,  which are  the image  of  the paramenter
     * points under a possibly unknown function.
     * @param np: the number of parameter points.
     * @param n: the first degree of this rectangular Bezier patch.
     * @param m: the second degree of this rectangular Bezier patch.
     * @param  rx: the  first coordinate  of  the first  point of  the
     * affine frame w.r.t. which this surface patch is defined.
     * @param  sx: the  second coordinate  of the  first point  of the
     * affine frame w.r.t. which this surface patch is defined.
     * @param  ry: the  first coordinate  of the  second point  of the
     * affine frame w.r.t. which this surface patch is defined.
     * @param sy:  the second  coordinate of the  second point  of the
     * affine frame w.r.t. which this surface patch is defined.
     *
     * @return A pointer to a newly created Bezier patch.
     *
     */
    static Pointer NEW(
                       double* pp,
                       double* sp,
                       unsigned np,
                       unsigned m,
                       unsigned n,
                       double rx,
                       double sx,
                       double ry,
                       double sy
                      )
    {
      tBezier *dumbPtr = new tBezier;
      Pointer smartPtr(dumbPtr);

      try {
        smartPtr->build( pp, sp, np, m, n, rx, sx, ry, sy );
      }
      catch (const tExceptionObject& xpt) {
        treatException(xpt);
        exit(0);
      }

      return smartPtr;
    }


    /**
     * Destructor.
     */
    ~tBezier();


    /**
     * Get a specific control point defining this Bezier patch.
     */
    t3DVector b(unsigned, unsigned) const throw (tExceptionObject);


    /**
     * Compute a point on this rectangular Bezier patch.
     */
    t3DVector point(double, double) const throw (tExceptionObject);

  private:
    /**
     * Default constructor
     */ 
    tBezier() {}


    /**
     * Creates  a  rectangular  Bezier  patch  from a  set  of  sample
     * points. The function uses  a least squares procedure to compute
     * the Bezier coefficients.
     *
     */
    void build(double*, double*, unsigned, unsigned, unsigned, double,
      double, double, double) throw (tExceptionObject);


    /**
     * Computes the "linear" index of a coefficient.
     *
     * @param i: first index of the coefficient.
     * @param j: second index of the coefficient.
     */
    unsigned index(unsigned i, unsigned j) const
    {
      return ( i * ( m_m + 1 ) ) + j;
    }


    /**
     * This function  computes the  value of one  Bernstein polynomial
     * (B_(n,i)) at a given parameter value with respect to the affine
     * frame [0,1].
     */
    double bernstein(unsigned, unsigned, double) const 
      throw(tExceptionObject);


    /**
     * This function  computes the value of  all Bernstein polynomials
     * at a  given parameter  value with respect  to the  affine frame
     * [0,1].
     */
    void all_bernstein(unsigned, double, std::vector<double>&) const 
      throw (tExceptionObject);


    /**
     * This function  computes the  pairwise product of  all Bernstein
     * polynomials of  order "m_m"  and "m_n" at  a set  of parameters
     * points.
     */
    void comp_bpoly_matrix(double**&, double*, unsigned np) const;


    /**
     * This function computes the product  of the transpose of a given
     * matrix with the matrix itself.
     */
    void comp_matrix_ata(double**, unsigned, unsigned, double**&)
      const;


    /**
     *  This function computes the product  of the transpose of a given
     *  (n x p) matrix "A" and a (n x 3) matrix B. The result is a (p x
     *  3) matrix.
     */
    void comp_matrix_atb(double**, double*, unsigned, unsigned,
      double**&) const;


    /**
     * Bi-degree of the Bezier patch.
     */
    unsigned m_m;
    unsigned m_n;


    /**
     * Affine  frame  with respect  to  which  this  surface patch  is
     * defined.
     */
    double m_rx;
    double m_sx;
    double m_ry;
    double m_sy;


    /**     
     * Control points.
     */
    double** m_b;

  };


}        // end of namespace pps


#endif
