/** 
 * @class dcel::tMyHEAttrib
 *  
 * @brief A class representing a set of half-edge attributes.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */


#ifndef MYHEATTRIB_H_
#define MYHEATTRIB_H_

#include "macros.h"
#include "exceptionobj.h"
#include "smartptr.h"
#include "dataobj.h"

#include "vertex.h"
#include "halfedge.h"
#include "edge.h"
#include "face.h"
#include "complex.h"
#include "heattrib.h"

#include "myvtattrib.h"

#include <stdlib.h>


#define HEATTRIB(h) \
        tSmartPtr<tMyHEAttrib>(dynamic_cast<tMyHEAttrib*> \
        (h->get_attrib().get_pointee()))


namespace pps {


  /**
   * Forward definition of classes.
   */
  struct tHalfEdge;

  
  class tMyHEAttrib : public dcel::tHEAttrib {
  public:
    /**
     * Defines Self as this class
     */    
    typedef tMyHEAttrib                 Self;


    /**
     * Defines Pointer as smart pointer to this class.
     */    
    typedef tSmartPtr<Self>             Pointer;


    /**
     * Defines ConstPointer as a const smart pointer to this class.
     */    
    typedef tSmartPtr<const Self>       ConstPointer;


    /**
     * Defines vtPointer as a smart pointer to a tVertex object.
     */
    typedef tSmartPtr<dcel::tVertex>    vtPointer;


    /**
     * Defines vaPointer as a smart pointer to a tMyVTAttrib object.
     */
    typedef tSmartPtr<pps::tMyVTAttrib> vaPointer;


    /**
     * Defines hePointer as a smart pointer to a tHalfEdge object.
     */
    typedef tSmartPtr<dcel::tHalfEdge>  hePointer;


    /**
     * Method for creating an instance of this object.
     *
     * @param h: a pointer to a half-edge.
     *
     */
    static Pointer NEW(const hePointer& h)
    {
      tMyHEAttrib *dumbPtr = new tMyHEAttrib;
      Pointer smartPtr(dumbPtr);

      smartPtr->set_owner(h);

      try {
        smartPtr->compute_id();
      }
      catch (const tExceptionObject& xpt) {
        treatException(xpt);
        exit(0);
      }

      return smartPtr;
    }


    /**
     * Destructor.
     */
    virtual ~tMyHEAttrib() {}


    /**
     * Returns the degree  of the origin vertex of  the half-edge that
     * owns this attribute.
     *
     * @return The degree  of the origin vertex of  the half-edge that
     * owns this attribute.
     *
     */
    unsigned get_degree() const;


    /**
     * Returns the ID of the owner of this attribute.
     *
     * @return The ID of the owner of this attribute.
     *
     */
    unsigned get_id() const { return m_id; }


    /**
     * Computes the  rectangular coordinates of  the image of  a point
     * under the transition map "g" associated with the half-edge that
     * owns this attribute.
     *
     */
    void g(double, double, double&, double&) const;


    /**
     * Computes the  rectangular coordinates of  the image of  a point
     * under  the   map  that  takes   the  point  to   the  canonical
     * domain.  This map is  associated with  the half-edge  that owns
     * this attribute.
     *
     */
    void to_cd(double, double, double&, double&) const;


    /**
     * Computes the  rectangular coordinates of  the image of  a point
     * under the map that takes the point from the canonical domain to
     * the  domain  associated  with  the  half-edge  that  owns  this
     * attribute.
     *
     */
    void from_cd(double, double, double&, double&) const;


  private:
    /**
     * Default constructor.
     */
    tMyHEAttrib() {}


    /**
     * Assigns a value to the ID of the owner of this attribute.
     *
     * @param  An unsigned integer  to be  assigned to  the ID  of the
     * owner of this attribute.
     *
     */
    void set_id(unsigned id) { m_id = id; }


    /**
     * Computes the ID of the half-edge that owns this attribute.
     *
     */
    void compute_id() throw (tExceptionObject);


    /**
     * Computes the location of a  given point after a rotation around
     * the origin by a given angle.
     */
    void rot_2d(double, double, double, double&, double&) const;


    /**
     * Id of the  half-edge with respect to the  set of halfedges that
     * share the origin vertex with this halfedge.
     */
    unsigned m_id;


  };


}


#endif   // MYHEATTRIB_H_
