/** 
 * @class dcel::tMyFCAttrib
 *  
 * @brief A class representing a set of half-edge attributes.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */


#ifndef MYFCATTRIB_H_
#define MYFCATTRIB_H_

#include "macros.h"
#include "exceptionobj.h"
#include "smartptr.h"
#include "dataobj.h"

#include "vertex.h"
#include "halfedge.h"
#include "edge.h"
#include "face.h"
#include "complex.h"
#include "fcattrib.h"

#include "3dvector.h"
#include "pntriangle.h"


#define FCATTRIB(f) \
        tSmartPtr<tMyFCAttrib>(dynamic_cast<tMyFCAttrib*> \
        (f->get_attrib().get_pointee()))


namespace pps {


  using common::t3DVector;

  
  class tMyFCAttrib : public dcel::tFCAttrib {
  public:
    /**
     * Defines Self as this class
     */    
    typedef tMyFCAttrib                Self;


    /**
     * Defines Pointer as smart pointer to this class.
     */    
    typedef tSmartPtr<Self>            Pointer;


    /**
     * Defines ConstPointer as a const smart pointer to this class.
     */    
    typedef tSmartPtr<const Self>      ConstPointer;


    /**
     * Defines fcPointer as a smart pointer to a tFace object.
     */
    typedef tSmartPtr<dcel::tFace>     fcPointer;


    /**
     * Defines pntPointer as a smart pointer to a tPNTriangle object.
     */
    typedef tSmartPtr<tPNTriangle>     pntPointer;


    /**
     * Method for creating an instance of this object.
     *
     * @param f: a pointer to a face.
     *
     */
    static Pointer NEW(const fcPointer& f)
    {
      tMyFCAttrib *dumbPtr = new tMyFCAttrib();
      Pointer smartPtr(dumbPtr);

      smartPtr->set_owner(f);

      smartPtr->set_normal(t3DVector(0,0,0));

      smartPtr->set_pnt(pntPointer(0));

      return smartPtr;
    }


    /**
     * Destructor.
     */
    virtual ~tMyFCAttrib() { m_pnt = pntPointer(0); }


    /**
     * Returns the unit normal of the supporting plane of this face.
     *
     * @return The unit normal of the supporting plane of this face.
     *
     */
     t3DVector get_normal() const { return m_n; }


    /**
     * Returns a pointer  to the PN triangle associated  with the face
     * that owns this attribute.
     *
     * @return A pointer  to the PN triangle associated  with the face
     * that owns this attribute.
     *
     */
     pntPointer get_pnt() const { return m_pnt; }


    /**
     * Assigns a vector to the  unit normal of the supporting plane of
     * the face that owns this attribute.
     *
     * @param  A vector  to  be assigned  to  the unit  normal of  the
     * supporting plane of the face that owns this attribute.
     *
     */
    void set_normal(const t3DVector& n) { m_n = n; }


    /**
     * Assigns a  value to the  pointer to the PN  triangle associated
     * with the  face that  owns this attribute.
     *
     * @param pnt: A pointer to a PN  triangle.
     *
     */
    void set_pnt(const pntPointer& pnt) { m_pnt = pnt; }


  private:
    /**
     * Default constructor.
     */
    tMyFCAttrib() {}


    /**
     * Unit normal of the supporting  plane of the face that owns this
     * attribute.
     */
    t3DVector m_n;


    /**
     * A pointer to the PN triangle associated with the face that owns
     * this attribute.
     */
    pntPointer m_pnt;

  };


}


#endif   // MYFCATTRIB_H_
