/** 
 * @class pps::tPNTriangle
 *  
 * @brief A class that implements PN triangles.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date Abril 2008
 *
 * @attention This program is distributed  WITHOUT ANY WARRANTY,  and
 *            it may be freely redistributed under the condition that
 *            the copyright notices are  not  removed,  and  no
 *            compensation  is  received. Private, research, and
 *            institutional use  is free. Distribution of this code as
 *            part of a commercial system is permissible ONLY BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */

#ifndef PNTRIANGLE_H_
#define PNTRIANGLE_H_

#include "smartptr.h"
#include "dataobj.h"
#include "macros.h"
#include "exceptionobj.h"

#include "3dvector.h"

#include <stdlib.h>


namespace pps {


  using common::t3DVector;


  /**
   * Class tPNTriangle: A class implementing a PN triangle.
   *
   */ 
  class tPNTriangle : public tDataObject {
  public:
    /**
     * Defines Self as this class.
     */
    typedef tPNTriangle           Self;


    /**
     * Defines Pointer as a smart pointer to this class.
     */
    typedef tSmartPtr<Self>       Pointer;

    
    /**
     * Defines ConstPointer as a const smart pointer to this class.
     */
    typedef tSmartPtr<const Self> ConstPointer;


    /**
     * Method for creating an instance of this object.
     *
     * @param p1: first vertex of the triangle defining the PN triangle.
     * @param p2: second vertex of the triangle defining the PN triangle.
     * @param p3: third vertex of the triangle defining the PN triangle.
     * @param n1: unit normal at the first vertex.
     * @param n2: unit normal at the second vertex.
     * @param n3: unit normal at the third vertex.
     *
     * @return A pointer to a newly created PN triangle.
     *
     */
    static Pointer NEW(
                       const t3DVector& p1, 
                       const t3DVector& p2, 
                       const t3DVector& p3,
                       const t3DVector& n1, 
                       const t3DVector& n2, 
                       const t3DVector& n3
                      )
    {
      tPNTriangle *dumbPtr = new tPNTriangle;
      Pointer smartPtr(dumbPtr);

      try {
        smartPtr->build(p1,p2,p3,n1,n2,n3);
      }
      catch (const tExceptionObject& xpt) {
        treatException(xpt);
        exit(0);
      }

      return smartPtr;
    }


    /**
     * Destructor.
     */
    ~tPNTriangle() {};


    /**
     * Computes a point on this PN triangle.
     */
    t3DVector point(double,double,double) const 
      throw (tExceptionObject);


    /**
     * Computes the unit normal of this PN triangle at a given point.
     */
    t3DVector normal(double,double,double) const 
      throw (tExceptionObject);


  private:
    /**
     * Default constructor.
     */
    tPNTriangle() {}


    /**
     * Computes the PN triangle coefficients
     */
    void build(const t3DVector&, const t3DVector&, const t3DVector&,
               const t3DVector&, const t3DVector&, const t3DVector&)
      throw (tExceptionObject);


    /**
     * The "b_(i,j,k)" coefficients of this PN triangle.
     */

    t3DVector m_b300;
    t3DVector m_b030;
    t3DVector m_b003;
    t3DVector m_b210;
    t3DVector m_b120;
    t3DVector m_b201;
    t3DVector m_b021;
    t3DVector m_b102;
    t3DVector m_b012;
    t3DVector m_b111;


    /**
     * The "n_(i,j,k)"  coefficients for computing the  unit normal of
     * this PN triangle at a  given point. These coefficients define a
     * quadratic interpolant.
     */

    t3DVector m_n200;
    t3DVector m_n020;
    t3DVector m_n002;
    t3DVector m_n110;
    t3DVector m_n101;
    t3DVector m_n011;
    
  };


}


#endif   // PNTRIANGLE_H_
