/** 
 * @class pps::tPPS
 *  
 * @brief A class implementing a parametric pseudo-surface.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date May 2008
 *
 * @attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "pps.h"

#include <sstream>

#include "macros.h"
#include "dcel.h"
#include "mathmacros.h"

#include "point.h"
#include "pntriangle.h"
#include "bezier.h"

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cassert>


namespace pps {


  using common::t3DVector;
  using dcel::tPoint;


  /**
   * Method for building a parametric pseudo-surface
   *
   *
   */
  void
  tPPS::build() throw (tExceptionObject)
  {
    /**
     * Check if the given complex is a simplicial surface.
     */
    if (!surf()->is_simplicial()) {
      std::stringstream ss(std::stringstream::in | 
                           std::stringstream::out);
      ss << "Input complex is not a simplicial surface.";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }


    /**
     * Check if the boundary of the given simplicial surface is empty.
     */
    if (surf()->has_boundary()) {
      std::stringstream ss(std::stringstream::in | 
                           std::stringstream::out);
      ss << "Simplicial surface must have an empty boundary.";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }


    /**
     * Create half-edge attributes.
     */
    create_hedge_attribs();  


    /**
     * Create face attributes.
     */
    create_face_attribs();  


    /**
     * Create vertex attributes.
     */
    create_vertex_attribs();  


    /**
     * Create a PN triangle surface from the simplicial surface.
     */
    create_pnt_surface();  


    /**
     * Create the parametrizations.
     */
    create_param();  

    /**
     * Computes the approximation error.
     */

  }


  /**
   * Method for creating half-edge attributes.
   *
   */
  void
  tPPS::create_hedge_attribs() const
  {
    tComplex::egIterator eit;
    for (eit = surf()->eg_begin(); eit != surf()->eg_end(); ++eit) {
      hePointer he1 = (*eit)->get_hedge();

      assert(he1.is_not_null());

      hePointer he2 = MATE(he1);

      assert(he2.is_not_null());

      haPointer at1 = tMyHEAttrib::NEW(he1);
      he1->set_attrib(at1);

      haPointer at2 = tMyHEAttrib::NEW(he2);
      he2->set_attrib(at2);
    }
  }


  /**
   * Method for creating face attributes.
   *
   */
  void
  tPPS::create_face_attribs() const
  {
    tComplex::fcIterator fit;
    for(fit = surf()->fc_begin(); fit != surf()->fc_end(); ++fit) {
      fcPointer fp = *fit;
      
      assert(fp.is_not_null());

      t3DVector n = comp_face_normal(fp);

      faPointer fa = tMyFCAttrib::NEW(fp);

      fa->set_normal(n); 

      fp->set_attrib(fa);
    }
  }


  /**
   * Method for creating vertex attributes.
   *
   */
  void
  tPPS::create_vertex_attribs() const
  {
    vtIterator vit;
    for(vit = surf()->vt_begin(); vit != surf()->vt_end(); ++vit) {
      vtPointer vp = *vit;

      t3DVector nv(0,0,0);
      unsigned i = 0;

      hePointer he1 = vp->get_hedge();
      hePointer he2 = he1;
      do {
        fcPointer fp = he2->get_face();
        faPointer fa = FCATTRIB(fp);

        assert(fa.is_not_null());

        nv += fa->get_normal();

        ++i;

        he2 = MATE(he2->prev());
      }
      while (he1 != he2);

      nv = (1 / double(i)) * nv;

      nv.normalize();

      vaPointer va = tMyVTAttrib::NEW(vp);

      va->set_normal(nv); 

      vp->set_attrib(va);
    }
  }


  /**
   * Method for creating a PN triangle surface.
   *
   */
  void
  tPPS::create_pnt_surface() const
  {
    tComplex::fcIterator fit;
    for (fit = surf()->fc_begin(); fit != surf()->fc_end(); ++fit) {
      fcPointer fp = *fit;

      hePointer he = fp->get_hedge();

      vtPointer v1 = he->get_vert();
      vtPointer v2 = he->next()->get_vert();
      vtPointer v3 = he->prev()->get_vert();   

      tPoint q1(v1->get_point());
      tPoint q2(v2->get_point());
      tPoint q3(v3->get_point());

      t3DVector p1(q1.x(),q1.y(),q1.z());
      t3DVector p2(q2.x(),q2.y(),q2.z());
      t3DVector p3(q3.x(),q3.y(),q3.z());

      vaPointer va1 = VTATTRIB(v1); 
      assert(va1.is_not_null());

      vaPointer va2 = VTATTRIB(v2); 
      assert(va2.is_not_null());

      vaPointer va3 = VTATTRIB(v3);
      assert(va3.is_not_null());

      t3DVector n1(va1->get_normal());
      t3DVector n2(va2->get_normal());
      t3DVector n3(va3->get_normal());

      tPNTriangle::Pointer pnt = 
         tPNTriangle::NEW(p1,p2,p3,n1,n2,n3);

      faPointer fa = FCATTRIB(fp);
      assert(fa.is_not_null());

      fa->set_pnt(pnt);
    }
  }


  /**
   * Method for creating the P-domain parametrizations.
   *
   */
  void
  tPPS::create_param() const
  {

    /**
     * For  each vertex  "v"  of the  simplicial  surface, sample  the
     * P-polygon   associated   with   it,   and  then   compute   the
     * corresponding points  in the PN triangle  surface. These points
     * are  then used  to compute  the parametrization  using  a least
     * squares procedure.
     */
    std::vector<double> pp;
    std::vector<double> sp;

    vtIterator vit;
    for (vit = surf()->vt_begin(); vit != surf()->vt_end(); ++vit) {
      /** 
       * Get the half-edge whose origin vertex is the current one.
       */
      vtPointer vp = *vit;

      /**
       * Sample the  p-domain associated  with the current  vertex, as
       * well  as  the PN  surface  patches  defined  on its  standard
       * triangulation.
       */
      try {
        build_patch( pp, sp, vp );
      }
      catch (const tExceptionObject& xpt) {
        treatException(xpt);
        exit(0);
      }
      
      /**
       * Release memory.
       */
      pp.clear();
      sp.clear();
    }
  }


  /**
   * Method for computing the unit normal of the supporting plane of a
   * given face.
   *
   * @param fp: pointer to a face.
   *
   */
  t3DVector
  tPPS::comp_face_normal(const fcPointer& fp) const 
  {
    tPoint p1 = fp->get_hedge()->get_vert()->get_point();
    tPoint p2 = fp->get_hedge()->next()->get_vert()->get_point();
    tPoint p3 = fp->get_hedge()->prev()->get_vert()->get_point();

    t3DVector v1(p1.x(),p1.y(),p1.z());
    t3DVector v2(p2.x(),p2.y(),p2.z());
    t3DVector v3(p3.x(),p3.y(),p3.z());

    t3DVector nu(v2 - v1);
    t3DVector nv(v3 - v1);

    return (nu.cross(nv)).normalize();
  }


  /**
   * Method  for sampling  a given  P-domain and  all the  PN triangle
   * surface  patches associated  with the  triangles of  the P-domain
   * standard triangulation.
   *
   * @param pp: array for storing the parameter points.
   * @param sp: array  for  storing  the points  on  the PN  triangle
   * surface.
   * @param vp: pointer to the vertex associated with the P-domain.
   *
   */
  void
  tPPS::build_patch(
                    std::vector<double>& pp,
                    std::vector<double>& sp,
                    const vtPointer& vp
                   ) 
    const throw (tExceptionObject)
  {
    assert( vp.is_not_null() );

    hePointer he = vp->get_hedge();

    assert( he.is_not_null() );

    haPointer ha = HEATTRIB(he);   

    assert( ha.is_not_null() );

    /**
     * Get the degree of the vertex associated with the P-domain.
     */
    unsigned nu = ha->get_degree();

    /**
     * Generate a  rectangular grid with (N  + 1) * (N  + 1) parameter
     * points.
     */
    const unsigned D = compute_patch_degree( nu );
    const unsigned N = compute_number_of_param_points( D );

    const double MYPI = acos(-1);
    const double R = cos ( MYPI / nu );

    /**
     * Coordinates of the lower leftmost point of the grid.
     */
    const double x0 = -R;
    const double y0 = -R;

    /**
     * Spacing  between  two  consecutives  points  in both  X  and  Y
     * directions.
     */
    const double dd = ( 2 * R ) / N;

    /**
     * Initialize the Y direction spacing counter.
     */   
    double dy = 0;

    unsigned j;
    for (j=0; j<=N; j++) {
      /** Y coordinate of the point. */
      double y = y0 + dy ;

      /**
       * Initialize the X direction spacing counter.
       */   
      double dx = 0 ;

      unsigned i;
      for (i=0; i<=N; i++) {
        /** X coordinate of the point. */
        double x = x0 + dx ;

        /**
         * If  this  parameter  point  is  inside  the  P-polygon  the
         * P-domain is  inscribed in, then  we can compute a  point on
         * the PN triangle surface.
         */

        double u;
        double v;
        double w;
        hePointer heaux = he;
        if ( find_triangle( x , y , heaux , u , v , w ) ) {
          /**
           * Store the coordinates of the parameter point.
           */
          pp.push_back( x );
          pp.push_back( y );

          /**
           * Compute  the  corresponding  point  in  the  PN  triangle
           * surface patch.
           */
          assert( heaux.is_not_null() );

          fcPointer fp = heaux->get_face();

          assert( fp.is_not_null() );   

          faPointer fa = FCATTRIB( fp );

          assert( fa.is_not_null() );

          t3DVector pt(0,0,0);

          if (heaux == heaux->get_face()->get_hedge()) {
            pt = fa->get_pnt()->point( u , v , w );
          }
          else if (heaux->next() == heaux->get_face()->get_hedge()) {
            pt = fa->get_pnt()->point( v , w , u );
          }
          else {
            pt = fa->get_pnt()->point( w , u , v );
          }

          /**
           * Store  the coordinates of  the point  on the  PN triangle
           * surface.
           */
          sp.push_back( pt[0] );
          sp.push_back( pt[1] );
          sp.push_back( pt[2] );
        }

        /** Increment X direction spacing counter. */
        dx += dd ;
      }

      /** Increment Y direction spacing counter. */
      dy += dd ;
    }

    /**
     * Create a rectangular Bezier patch.
     */

    tBezier::Pointer pat = 
      tBezier::NEW( &pp[0], &sp[0], pp.size() >> 1, D , D, -R , R, -R , R );

    vaPointer va = VTATTRIB(vp);

    assert( va.is_not_null() );

    va->set_patch( pat );

    return;
  }


  /**
   * Method for finding the  triangle of the standard triangulation of
   * a  P-domain that  contains  a given  point  (if there  is such  a
   * triangle).
   *
   * @param x: first rectangular coordinate of the given point.
   * @param y: second rectangular coordinate of the given point.
   * @param  he:  a  half-edge  whose  origin  vertex  is  the  vertex
   * associated with  the P-domain. This  parameter is also  an output
   * parameter, as  it will be  the half-edge of the  "found" triangle
   * with origin in the same vertex.
   * @param u:  first barycentric coordinate  of the given  point with
   * respect to the triangle that contains it (if any).
   * @param v:  second barycentric coordinate  of the given  point with
   * respect to the triangle that contains it (if any).
   * @param w:  third barycentric coordinate  of the given  point with
   * respect to the triangle that contains it (if any).
   *
   * @return True if a triangle is found and false otherwise.
   *
   */
   bool
   tPPS::find_triangle(
                      double x,
                      double y,
                      hePointer& he,
                      double& u,
                      double& v,
                      double& w
		     )
    const
  {
    hePointer heaux = he;
    do {
      assert( he.is_not_null() );

      haPointer ha = HEATTRIB(heaux);

      assert( ha.is_not_null() );

      /**
       * Get the degree of the vertex associated with the P-domain.
       */
      unsigned nu = ha->get_degree();

      /**
       * Get the degree of the current half-edge.
       */
      unsigned id = ha->get_id();

      const double MYPI = acos(-1);
      const double ang  = 2 * ( MYPI / nu );

      double x1 = cos(   id       * ang );
      double y1 = sin(   id       * ang );
      double x2 = cos( ( id + 1 ) * ang );
      double y2 = sin( ( id + 1 ) * ang );

      try {
        comp_bary_coords(0, 0, x1, y1, x2, y2, x, y, u, v, w);  
      }
      catch (const tExceptionObject& xpt) {
        treatException(xpt);
        exit(0);
      }

      if ( 
           ( u >= 0 ) && ( u <= 1 ) && ( v >= 0 ) && 
           ( v <= 1 ) && ( w >= 0 ) && ( w <= 1 )
	 ) 
      {
        he = heaux;

        return true;
      }
      else {
        heaux = MATE( heaux->prev() ) ;
      } 
    }
    while ( heaux != he );

    return false;
  }


  /**
   * Method for computing the approximation error.
   *
   */
  void
  tPPS::comp_approx_error()
  {
    /**
     * For  each vertex  "v"  of the  simplicial  surface, sample  the
     * P-polygon associated with  it, compute the corresponding points
     * in the PN triangle surface and in the PPS, and then compute the
     * Euclidean distance between the points corresponding to the same
     * parameter point.  The Euclidean distance is used to compute the
     * maximum approximation  error and  the RMS of  the approximation
     * error.
     */
    std::vector<double> pp;
    std::vector<double> sp1;
    std::vector<double> sp2;

    vtIterator vit;
    for (vit = surf()->vt_begin(); vit != surf()->vt_end(); ++vit) {
      /** 
       * Get the half-edge whose origin vertex is the current one.
       */
      vtPointer vp = *vit;

      /**
       * Sample the parameter space.
       */
      sample_param_space( vp, pp );

      try {
        /**
         * Compute points on the PN triangle surface.
         */
        sample_pnt_surface( vp, pp, sp1 ); 

        /**
         * Compute points on the PPS triangle surface.
         */
        sample_pps( vp, pp, sp2 );
      }
      catch (const tExceptionObject& xpt) {
        treatException(xpt);
        exit(0);
      }

      /**
       * Compute the maximum and approximation error and its RMS.
       */
      const unsigned N = pp.size() >> 1 ;

      assert( N > 0 );

      double sdmax = 0;
      double sdmin = 0;
      double sdtot = 0;

      unsigned i;
      for (i=0; i<N; i++) {
        /**
         * Get  the  coordinates  of  the  point on  the  PN  triangle
         * surface.
         */
        double x1 = sp1[ 3 * i     ];
        double y1 = sp1[ 3 * i + 1 ];
        double z1 = sp1[ 3 * i + 2 ];

        /**
         * Get  the  coordinates  of  the  point on  the  PPS.
         */
        double x2 = sp2[ 3 * i     ];
        double y2 = sp2[ 3 * i + 1 ];
        double z2 = sp2[ 3 * i + 2 ];

        /**
         * Compute the square of the Euclidean distance.
         */
        double dx = ( x1 - x2 );
        double dy = ( y1 - y2 );
        double dz = ( z1 - z2 );
        double sd = ( dx * dx ) + ( dy * dy ) + ( dz * dz ) ;
 
        if (sd > sdmax ) {
          sdmax = sd;
        }

        if ( i == 0 ) {
          sdmin = sd;
        }
        else if ( sd < sdmin ) {
          sdmin = sd;
        }

        sdtot += sd ;
      }

      /**
       * Compute the root mean square.
       */
      set_rms_approx_error( sqrt( sdtot / N ) );

      /**
       * Compute the maximum approximation error.
       */ 
      set_max_approx_error( sqrt( sdmax ) );


      /**
       * Compute the minimum approximation error.
       */ 
      set_min_approx_error( sqrt( sdmin ) );

      /**
       * Release memory.
       */
      pp.clear();
      sp1.clear();
      sp2.clear();
    }

    return;
  }


  /**
   * Method  for sampling  a given  P-domain.
   *
   * @param vp: pointer to the vertex associated with the P-domain.
   * @param pp: array for storing the parameter points.
   *
   */
  void
  tPPS::sample_param_space(
                           const vtPointer& vp,
                           std::vector<double>& pp
                          ) 
    const
  {
    assert( vp.is_not_null() );

    vaPointer va = VTATTRIB( vp );

    assert( va.is_not_null() );

    /**
     * Get the degree of the vertex associated with the P-domain.
     */
    unsigned nu = va->get_degree();

    /**
     * Generate a  rectangular grid with (N  + 1) * (N  + 1) parameter
     * points.
     */
    const unsigned D = compute_patch_degree( nu );
    const unsigned N = compute_number_of_param_points( D );

    const double MYPI = acos(-1);
    const double R    = cos( MYPI / nu );

    /**
     * Coordinates of the lower leftmost point of the grid.
     */
    const double x0 = -R;
    const double y0 = -R;

    /**
     * Spacing  between  two  consecutives  points  in both  X  and  Y
     * directions.
     */
    const double dd = ( 2 * R ) / N;

    /**
     * Initialize the Y direction spacing counter.
     */   
    double dy = 0;

    unsigned j;
    for (j=0; j<=N; j++) {
      /** Y coordinate of the point. */
      double y = y0 + dy ;

      /**
       * Initialize the X direction spacing counter.
       */   
      double dx = 0 ;

      unsigned i;
      for (i=0; i<=N; i++) {
        /** X coordinate of the point. */
        double x = x0 + dx ;

        /**
         * If this  parameter point is inside  the p-domain associated
         * with "vp" then store it in the array "pp".
         */
        if ( ( ( x * x ) + ( y * y ) ) < ( R * R ) ) {
          /**
           * Store the coordinates of the parameter point.
           */
          pp.push_back( x );
          pp.push_back( y );
        }

        /** Increment X direction spacing counter. */
        dx += dd ;
      }

      /** Increment Y direction spacing counter. */
      dy += dd ;
    }

    return;
  }


  /**
   * Method for sampling the PN triangle surface with parameter points
   * defined in a given P-domain.
   *
   * @param vp: pointer to the vertex associated with the P-domain.
   * @param pp: array for storing the parameter points.
   * @param  sp: array  for  storing  the points  on  the PN  triangle
   * surface.
   *
   */
  void
  tPPS::sample_pnt_surface(
                           const vtPointer& vp,
                           std::vector<double>& pp,
                           std::vector<double>& sp
                          ) 
    const throw (tExceptionObject)
  {
    assert( vp.is_not_null() );

    hePointer he = vp->get_hedge();

    assert( he.is_not_null() );

    /**
     * Get the number of parameter points.
     */
    unsigned N = pp.size() >> 1 ;

    /**
     * For  each  parameter point,  compute  its  image  under the  PN
     * triangle surface.
     */
    unsigned i;
    for (i=0; i<N; i++) {
      /**
       * Get the coordinates of the parameter point.
       */
      double xp = pp[ (i << 1)     ];
      double yp = pp[ (i << 1) + 1 ];

      /**
       * Find the  triangle that contains  the parameter point  in the
       * canonical triangulation associated with vertex "vp".
       */
      double u;
      double v;
      double w;
      hePointer heaux = he;
      if ( find_triangle( xp , yp , heaux , u , v , w ) ) {
        /**
         * Compute the corresponding point  in the PN triangle surface
         * patch.
         */
        fcPointer fp = heaux->get_face();

        assert( fp.is_not_null() );   

        faPointer fa = FCATTRIB( fp );

        assert( fa.is_not_null() );

        t3DVector pt(0,0,0);

        if (heaux == heaux->get_face()->get_hedge()) {
          pt = fa->get_pnt()->point( u , v , w );
        }
        else if (heaux->next() == heaux->get_face()->get_hedge()) {
          pt = fa->get_pnt()->point( v , w , u );
        }
        else {
          pt = fa->get_pnt()->point( w , u , v );
        }

        sp.push_back( pt.m_x );
        sp.push_back( pt.m_y );
        sp.push_back( pt.m_z );
      }
      else {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Numerical instability in the point in triangle test!";
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }
    }

    return;
  }


  /**
   * Method for  sampling the PPS  with parameter points defined  in a
   * given P-domain.
   *
   * @param vp: pointer to the vertex associated with the P-domain.
   * @param pp: array for storing the parameter points.
   * @param  sp: array  for  storing  the points  on  the PPS.
   *
   */
  void
  tPPS::sample_pps(
                   const vtPointer& vp,
                   std::vector<double>& pp,
                   std::vector<double>& sp
                  ) 
    const throw (tExceptionObject)
  {
    assert( vp.is_not_null() );

    hePointer he = vp->get_hedge();

    assert( he.is_not_null() );

    /**
     * Get the number of parameter points.
     */
    unsigned N = pp.size() >> 1 ;

    /**
     * For each parameter point, compute its image under the PPS.
     */
    unsigned i;
    for (i=0; i<N; i++) {
      /**
       * Get the coordinates of the parameter point.
       */
      double xp = pp[ (i << 1)     ];
      double yp = pp[ (i << 1) + 1 ];

      /**
       * Find the  triangle that contains  the parameter point  in the
       * canonical triangulation associated with vertex "vp".
       */
      double u;
      double v;
      double w;
      hePointer heaux = he;
      if ( find_triangle( xp , yp , heaux , u , v , w ) ) {
        /**
         * Compute the corresponding point  in the PPS patch aaociated
         * with the triangle defined by "heaux".
         */

        t3DVector pt( this->eval( u, v, w , heaux ) );

        sp.push_back( pt.m_x );
        sp.push_back( pt.m_y );
        sp.push_back( pt.m_z );
      }
      else {
        std::stringstream ss (std::stringstream::in | 
                              std::stringstream::out);
        ss << "Numerical instability in the point in triangle test!";
        throw tExceptionObject(__FILE__,__LINE__,ss.str());
      }
    }

    return;
  }


  /**
   * General  parametric   function  defining  this  pseudo-parametric
   * surface  at a  point given  by its  barycentric  coordinates with
   * respect to the affine  frame (triangle) (0,0), (1,0), (SQRT(3)/2,
   * 1/2).
   *
   * @param u: first barycentric coordinate.
   * @param v: second barycentric coordinate.
   * @param w: third barycentric coordinate.
   * @param f: a  face of the simplicial surface  associated with this
   * PPS.
   * @param p: a vector to store  the point in 3D corresponding to the
   * image of  the parameter point whose  barycentric coordinates were
   * given.
   *
   */
  void
  tPPS::eval(
             double u,
             double v,
             double w,
             const fcPointer& f,
             double p[3]
            )
    const throw (tExceptionObject)
  {
    /**
     * Check whether the iterator points to a valid face.
     */
    if (f.is_null()) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Invalid reference to surface mesh face!";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    try {
      /**
       * Compute a point on the PPS.
       */
      t3DVector pt( this->eval( u, v, w , f->get_hedge() ) );

      p[0] = pt.m_x;
      p[1] = pt.m_y;
      p[2] = pt.m_z;
    }
    catch (const tExceptionObject& xpt) {
      treatException(xpt);
      exit(0);
    }
  }


  /**
   * Auxiliary   method   implementing   a  function   defining   this
   * pseudo-parametric  surface at  a point  given by  its barycentric
   * coordinates with  respect to  the affine frame  (triangle) (0,0),
   * (1,0), (SQRT(3)/2, 1/2).
   *
   * @param u: first barycentric coordinate.
   * @param v: second barycentric coordinate.
   * @param w: third barycentric coordinate.
   * @param f: a  face of the simplicial surface  associated with this
   * PPS.
   * @param p: a vector to store  the point in 3D corresponding to the
   * image of  the parameter point whose  barycentric coordinates were
   * given.
   *
   */
  t3DVector
  tPPS::eval(
             double u,
             double v,
             double w,
             const hePointer& he
            )
    const throw (tExceptionObject)
  {
    assert( ( u >= 0 ) && ( u <= 1 ) );
    assert( ( v >= 0 ) && ( v <= 1 ) );
    assert( ( w >= 0 ) && ( w <= 1 ) );
    assert( ABS(1 - ( u + v + w ) ) <= 1e-15 ) ;

    assert(he.is_not_null());

    hePointer he1 = he ;

    /**
     * Map  the point  with barycentric  coordinates (u,v,w)  from the
     * canonical  domain to  the P-domain  associated with  the origin
     * vertex of the half-edge "he".
     */
    double xc, yc;
    from_bary_to_rect(u, v, w, 0, 0, 1, 0, 0.5, 0.5*sqrt(3), xc, yc);

    haPointer ha = HEATTRIB(he1);

    assert(ha.is_not_null());

    double xr, yr;
    ha->from_cd(xc, yc, xr, yr);

    const double MYPI = acos(-1);

    unsigned nu = ha->get_degree();

    const double R = cos( MYPI / nu ); 

    /**
     * If the  point is  not inside  the P-domain, we  do not  have to
     * proceed with the calculation of the contribution of the current
     * P-domain.
     *
     */
    double ll = ( xr * xr ) + ( yr * yr ) ;

    double sw = 0;
    t3DVector sf( 0 , 0 , 0 );    

    /**
     * If the current domain  contains the point (xr,yr), computes its
     * contribution.
     *
     */
    if ( ll < R * R ) {
      /** Compute the value of the weight function */
      sw = this->wfunc( xr , yr , R );

      /** Compute the value of the shape function */
      vaPointer va = VTATTRIB( he1->get_vert() );

      assert( va.is_not_null() );

      assert( va->get_patch().is_not_null() );

      t3DVector fval( va->get_patch()->point( xr , yr ) );

      sf = sw * fval ;
    }

    if ( ( u != 1 ) && ( v != 1 ) && ( w != 1 ) ) {
      /**
       * If the point  lies on an edge, then we  can consider only two
       * P-domains.
       *
       */
      if ( w == 0 ) {
        double ur;
        double vr;
        ha->g( xr , yr , ur , vr );

        try {
          comp_lpd_contrib( MATE( he1 ) , ur , vr , sw , sf );
        }
        catch (const tExceptionObject& xpt) {
          treatException(xpt);
          exit(0);
        }
      }
      else if ( v == 0 ) {
        double ur;
        double vr;

        hePointer he2 = MATE( he1->prev() );

        assert( he2.is_not_null() );

        haPointer ha2 = HEATTRIB( he2 );

        assert( ha2.is_not_null() );

        ha2->g( xr , yr , ur , vr );

        try {
          comp_rpd_contrib( MATE( he2 ) , ur , vr , sw , sf );
        }
        catch (const tExceptionObject& xpt) {
          treatException(xpt);
          exit(0);
        }
      }
      else {
        double ur;
        double vr;
        ha->g( xr , yr , ur , vr );

        try {
          comp_lpd_contrib( MATE( he1 ), ur, vr, sw, sf );
        }
        catch (const tExceptionObject& xpt) {
          treatException(xpt);
          exit(0);
        }

        hePointer he2 = MATE( he1->prev() );
        haPointer ha2 = HEATTRIB( he2 );

        assert( ha2.is_not_null() );

        ha2->g( xr , yr , ur , vr );

        try {
          comp_rpd_contrib( MATE( he2 ) , ur , vr , sw , sf );
        }
        catch (const tExceptionObject& xpt) {
          treatException(xpt);
          exit(0);
        }
      }
    }

    assert( sw != 0 ) ;

    return (1./sw) * sf;
  }


  /**
   * General parametric  function defining a PN triangle  surface at a
   * point given  by its barycentric  coordinates with respect  to the
   * affine  frame  (triangle)  (0,0),  (cos(pi/6),  -sin(pi/6)),  and
   * (cos(pi/6), sin(pi/6)).
   *
   * @param u: first barycentric coordinate.
   * @param v: second barycentric coordinate.
   * @param w: third barycentric coordinate.
   * @param f: a  face of the simplicial surface  associated with this
   * PPS.
   *
   * @return A point in 3D corresponding to the image of the parameter
   * point whose barycentric coordinates were given.
   */
  t3DVector
  tPPS::eval_pnt(
                 double u,
                 double v,
                 double w,
                 const fcPointer& f
                )
    const
  {
    /**
     * Check whether the iterator points to a valid face.
     */

    assert(f.is_not_null());

    faPointer fa = FCATTRIB(f);

    assert(fa.is_not_null());

    return fa->get_pnt()->point(u,v,w);
  }


  /**
   * Computes the contribution of the surface patch associated a given
   * P-domain at a given point.  The P-domain is implicitly identified
   * by  a  given  half-edge   (whose  origin  vertex  is  the  vertex
   * associated with the P-domain). The  given point is expected to be
   * in the P-triangle on the left of the given half-edge.
   *
   * @param he: pointer to a given half-edge.
   * @param xr: first rectangular coordinate of the given point.
   * @param yr: second rectangular coordinate of the given point.
   * @param sw: accumulator of the sum of the weight functions.
   * @param sf: accumulator of the sum  of parametrizations.
   *
   */
  void
  tPPS::comp_lpd_contrib(
                         const hePointer& he1,
                         double xr,
                         double yr,
                         double& sw,
                         t3DVector& sf
		        )
    const throw (tExceptionObject)
  {
    assert( he1.is_not_null() );

    haPointer ha1 = HEATTRIB( he1 );

    const double MYPI = acos(-1);

    unsigned nu = ha1->get_degree();

    const double R = cos( MYPI / nu ); 

    /**
     * If the point (xr,yr) is  inside the P-domain associate with the
     * origin    vertex    of    "he1",    then   we    compute    its
     * contribution. Otherwise, we ignore it.
     */
    double ll = ( xr * xr ) + ( yr * yr ) ;

    if ( ll < R * R ) {
      /** Compute the value of the weight function. */

      double w = this->wfunc( xr , yr , R );
      sw += w;

      vaPointer va1 = VTATTRIB( he1->get_vert() );

      assert( va1.is_not_null() );

      assert( va1->get_patch().is_not_null() );

      t3DVector fval( va1->get_patch()->point( xr , yr ) );

      sf += (w * fval);
    }

    return;
  }


  /**
   * Computes the contribution of the surface patch associated a given
   * P-domain at a given point.  The P-domain is implicitly identified
   * by  a  given  half-edge   (whose  origin  vertex  is  the  vertex
   * associated with the P-domain). The  given point is expected to be
   * in the P-triangle on the right of the given half-edge.
   *
   * @param he: pointer to a given half-edge.
   * @param xr: first rectangular coordinate of the given point.
   * @param yr: second rectangular coordinate of the given point.
   * @param sw: accumulator of the sum of the weight functions.
   * @param sf: accumulator of the sum  of parametrizations.
   *
   */
  void
  tPPS::comp_rpd_contrib(
                         const hePointer& he1,
                         double xr,
                         double yr,
                         double& sw,
                         t3DVector& sf
		        )
    const throw (tExceptionObject)
  {
    assert( he1.is_not_null() );

    haPointer ha1 = HEATTRIB( he1 );

    const double MYPI = acos(-1);

    unsigned nu = ha1->get_degree();

    const double R = cos( MYPI / nu ); 

    /**
     * If the point (xr,yr) is  inside the P-domain associate with the
     * origin    vertex    of    "he1",    then   we    compute    its
     * contribution. Otherwise, we ignore it.
     */
    double ll = ( xr * xr ) + ( yr * yr ) ;

    if ( ll < R * R ) {
      /** Compute the value of the weight function. */
      double w = this->wfunc( xr , yr , R );
      sw += w;

      vaPointer va1 = VTATTRIB( he1->get_vert() );

      assert( va1.is_not_null() );

      assert( va1->get_patch().is_not_null() );

      t3DVector fval( va1->get_patch()->point( xr , yr ) );

      sf += (w * fval);
    }

    return;
  }


  /**
   * Converts the barycentric coordinates  of a point in the reference
   * triangle to the corresponding rectangular coordinates.
   *
   * @param u: 1st barycentric coordinate of the point. 
   * @param v: 2nd barycentric coordinate of the point.
   * @param w: 3rd barycentric coordinate of the point.
   * @param  x1: X  coordinate  of  the 1st  vertex  of the  reference
   * triangle.
   * @param  y1: Y  coordinate  of  the 1st  vertex  of the  reference
   * triangle.
   * @param  x2: X  coordinate  of  the 2nd  vertex  of the  reference
   * triangle.
   * @param  y2: Y  coordinate  of  the 2nd  vertex  of the  reference
   * triangle.
   * @param  x3: X  coordinate  of  the 3rd  vertex  of the  reference
   * triangle.
   * @param  y3: Y  coordinate  of  the 3rd  vertex  of the  reference
   * triangle.
   * @param x: 1st rectangular coordinate of the resulting point.
   * @param y: 2nd rectangular coordinate of the resulting point.
   *
   */
  void
  tPPS::from_bary_to_rect(
                          double u,
                          double v,
                          double w,
                          double x1,
                          double y1,
                          double x2,
                          double y2,
                          double x3,
                          double y3,
                          double& x,
                          double& y
			 )
    const
  {
    x = (u * x1) + (v * x2) + (w * x3);
    y = (u * y1) + (v * y2) + (w * y3);
  }


  /**
   * Computes  the  barycentric  coordinates  of  a  given  point  (in
   * rectangular  coordinates)  with  respect  to  a  given  reference
   * triangle
   *
   * @param x1: first coordinate of  the first vertex of the reference
   * triangle.  
   * @param y1: second coordinate of the first vertex of the reference
   * triangle.
   * @param x2: first coordinate of the second vertex of the reference
   * triangle.
   * @param  y2:  second  coordinate  of  the  second  vertex  of  the
   * reference triangle.
   * @param x3: first coordinate of  the third vertex of the reference
   * triangle.
   * @param y3: second coordinate of the third vertex of the reference
   * triangle.
   * @param xp: first rectangular coordinate of the point.
   * @param yp: second rectangular coordinate of the point.
   * @param u: first barycentric coordinate of the point.
   * @param v: second barycentric coordinate of the point.
   * @param w: third barycentric coordinate of the point.
   *
   */
  void
  tPPS::comp_bary_coords(
                         double x1,
                         double y1,
                         double x2,
                         double y2,
                         double x3,
                         double y3,
                         double xp,
                         double yp,
                         double& u,
                         double& v,
                         double& w
			)
    const throw (tExceptionObject)
  {
    double dd = (x2 * y1) - (x3 * y1) - (x1 * y2)
              + (x3 * y2) + (x1 * y3) - (x2 * y3);

    if (ISZERO(dd)) {
      std::stringstream ss (std::stringstream::in | 
                            std::stringstream::out);
      ss << "Determinant is nearly zero!";
      throw tExceptionObject(__FILE__,__LINE__,ss.str());
    }

    u = (x3 * y2) - (xp * y2) - (x2 * y3) 
      + (xp * y3) + (x2 * yp) - (x3 * yp);

    u /= dd;

    v = (xp * y1) - (x3 * y1) + (x1 * y3)
      - (xp * y3) - (x1 * yp) + (x3 * yp);

    v /= dd;

    if (ABS(u) < 1e-14) {
      u = 0;
    }
    else if (ABS(1-u) < 1e-14) {
      u = 1;
    }

    if (ABS(v) < 1e-14) {
      v = 0;
    }
    else if (ABS(1-v) < 1e-14) {
      v = 1;
    }

    w = 1 - u - v;

    if (ABS(w) < 1e-14) {
      w = 0;
    }
    else if (ABS(1-w) < 1e-14) {
      w = 1;
    }
  }


  /**
   * Evaluates the weight function at a given point.
   *
   * @param  u: 1st barycentric coordinate of the given point.
   * @param  v: 2nd barycentric coordinate of the given point.
   * @param  w: 3rd barycentric coordinate of the given point.
   * @param x2:  X coordinate  of  the  3rd  vertex of  the  triangle
   * corresponding to  the support of this weight  function (the other
   * two points are (0,0) and (1,0)).
   * @param y2:  Y coordinate  of  the  3rd  vertex of  the  triangle
   * corresponding to  the support of this weight  function (the other
   * two points are (0,0) and (1,0)).
   * @param  lb: lower  bound for  the  "zero" region  of this  weight
   * function.
   * @param  ub: upper  bound for  the  "zero" region  of this  weight
   * function.
   *
   * @return the function value at the given point.
   *
   */
  double
  tPPS::wfunc(
              double x,
              double y,
              double R
             ) 
    const
  {
    double ll = sqrt( x * x + y * y );

    return nfunc( ll , 0.25 * R , R );
  }


  /**
   * Evaluates the "eta" blend function at a given parameter value
   *
   * @param  s: parameter value.
   * @param d1: lower bound for the blending region.
   * @param d2: upper bound for the blending region.
   *
   * @return The function value at "s".
   *
   */
  double
  tPPS::nfunc( 
              double s,
              double d1,
              double d2
             ) 
    const
  {
    assert( d2 > d1 );
    assert( d1 >  0 );
    assert( d2 <  1 );

    double res;

    if (s <= d1) {
      res = 1;
    }
    else if (s < d2) {
      double h1 = ( s - d1 ) / ( d2 - d1 );
      double h2 = 1 / sqrt( 1 - h1 );

      h1 = 1 / sqrt(     h1 );
      h2 = exp( h2 - h1 );

      res = 1 / ( 1 + ( h2 * h2 ) );
    }
    else {
      res = 0;
    }

    return res;
  }


  /**
   * Samples this  PPS and  creates a "surface  mesh" with  the sample
   * points.   This  method is  meant  to  be  used for  visualization
   * purposes only.
   *
   * @param lod: level of detail of the sampling.
   * @param lv1: list of points of the PPS surface mesh.
   * @param lv2: list of points of the PN triangle surface mesh.
   * @param lf: list of faces of the surface mesh.
   *
   */
  void
  tPPS::sample(
               unsigned lod,
               std::vector<double>& lv1,
               std::vector<double>& lv2,
               std::vector<unsigned>& lf
	      )
  {
    /**
     * Assign an ID to each vertex of the surface mesh.
     */
    vtIterator vit;
    unsigned i = 0;
    for (vit = surf()->vt_begin(); vit != surf()->vt_end(); ++vit) {

      vtPointer vt = *vit;

      m_vnt.insert( std::make_pair( vt, i ) );

      /**
       * Compute the position of a vertex of the PPS surface mesh.
       */
      t3DVector pt(eval(1,0,0,(*vit)->get_hedge()));

      /**
       * Store the vertex in the set of vertices.
       */
      lv1.push_back(pt.m_x);
      lv1.push_back(pt.m_y);
      lv1.push_back(pt.m_z);

      hePointer he = (*vit)->get_hedge();

      if (he == he->get_face()->get_hedge()) {
        pt = eval_pnt(1,0,0,he->get_face());
      }
      else if (he->next() == he->get_face()->get_hedge()) {
        pt = eval_pnt(0,0,1,he->get_face());
      }
      else {
        pt = eval_pnt(0,1,0,he->get_face());
      }

      // Store the vertex in the set of vertices.
      lv2.push_back(pt.m_x);
      lv2.push_back(pt.m_y);
      lv2.push_back(pt.m_z);

      ++i;      
    }

    /**
     * Sample each  face using  a "midpoint" subdivision  (like Loop's
     * scheme), and then connect the points to form the faces. We keep
     * the points and faces in a list.
     */
    fcIterator fit;
    for (fit = surf()->fc_begin(); fit != surf()->fc_end(); ++fit) {
      midpointsub(lod,i,*fit,lv1,lv2,lf);
    }

    m_vnt.clear();
    m_vpa.clear();
  }


  /**
   * Samples a "curved face" of this surface.
   *
   * @param lod: level of detail of the sampling.
   * @param vcount: vertex counter.
   * @param f: pointer to a face of the PPS.
   * @param lv1: list of points of the PPS surface mesh.
   * @param lv2: list of points of the PN triangle surface mesh.
   * @param lf: list of faces of the surface mesh.
   */
  void
  tPPS::midpointsub(
                    unsigned lod,
                    unsigned& vcount,
                    const fcPointer& f,
                    std::vector<double>& lv1,
                    std::vector<double>& lv2,
                    std::vector<unsigned>& lf
	           )
  {
    /**
     * Get the vertices of face "f".
     */
    vtPointer v1 = f->get_hedge()->get_vert();
    vtPointer v2 = f->get_hedge()->next()->get_vert();
    vtPointer v3 = f->get_hedge()->prev()->get_vert();

    /**
     * Get the ID of the above vertices.
     */
    unsigned id1 = m_vnt.find(v1)->second;
    unsigned id2 = m_vnt.find(v2)->second;
    unsigned id3 = m_vnt.find(v3)->second;

    /**
     * If  we already reached  the largest  LOD then  we can  create a
     * face.
     */
    if (lod == 0) {
      lf.push_back(id1);
      lf.push_back(id2);
      lf.push_back(id3);
    }
    else {
      t3DVector c1(1,0,0);
      t3DVector c2(0,1,0);
      t3DVector c3(0,0,1);

      midpointsub(lod-1,vcount,f,id1,id2,id3,c1,c2,c3,lv1,lv2,lf);
    }
  }


  /**
   *
   * Samples a "curved face" of this surface.
   *
   * @param lod: level of detail of the sampling.
   * @param vcount: vertex counter.
   * @param f: pointer to a face of the PPS.
   * @param v1: ID of the first vertex of the face.
   * @param v2: ID of the second vertex of the face.
   * @param v3: ID of the third vertex of the face. 
   * @param  c1: barycentric  coordinate of  the first  vertex  of the
   * face.
   * @param  c2: barycentric coordinate  of the  second vertex  of the
   * face.
   * @param  c3: barycentric  coordinate of  the third  vertex  of the
   * face.
   * @param lv1: list of points of the PPS surface mesh.
   * @param lv2: list of points of the PN triangle surface mesh.
   * @param lf: list of faces of the surface mesh.
   *
   */
  void
  tPPS::midpointsub(
                    unsigned lod,
                    unsigned& vcount,
                    const fcPointer& f,
                    unsigned v1,
                    unsigned v2,
                    unsigned v3,
                    const t3DVector& c1,
                    const t3DVector& c2,
                    const t3DVector& c3,
                    std::vector<double>& lv1,
                    std::vector<double>& lv2,
                    std::vector<unsigned>& lf
	           )
  {
    /**
     * Compute the midpoint vertices if they do not exist already.
     */
    t3DVector c12(0.5 * (c1 + c2));
    t3DVector c23(0.5 * (c2 + c3));
    t3DVector c31(0.5 * (c3 + c1));

    VPATAB::iterator pit = m_vpa.find(std::make_pair(v1,v2));

    unsigned v12;
    if (pit != m_vpa.end()) {
      v12 = pit->second;
    }
    else {
      t3DVector pos = eval(c12.m_x,c12.m_y,c12.m_z,f->get_hedge());
      lv1.push_back(pos.m_x);
      lv1.push_back(pos.m_y);
      lv1.push_back(pos.m_z);

      pos = eval_pnt(c12.m_x,c12.m_y,c12.m_z,f);
      lv2.push_back(pos.m_x);
      lv2.push_back(pos.m_y);
      lv2.push_back(pos.m_z);

      v12 = vcount;
      m_vpa.insert(std::make_pair(std::make_pair(v2,v1),v12));
      ++vcount;
    }

    pit = m_vpa.find(std::make_pair(v2,v3));

    unsigned v23;
    if (pit != m_vpa.end()) {
      v23 = pit->second;
    }
    else {
      t3DVector pos = eval(c23.m_x,c23.m_y,c23.m_z,f->get_hedge());
      lv1.push_back(pos.m_x);
      lv1.push_back(pos.m_y);
      lv1.push_back(pos.m_z);

      pos = eval_pnt(c23.m_x,c23.m_y,c23.m_z,f);
      lv2.push_back(pos.m_x);
      lv2.push_back(pos.m_y);
      lv2.push_back(pos.m_z);

      v23 = vcount;
      m_vpa.insert(std::make_pair(std::make_pair(v3,v2),v23));
      ++vcount;
    }

    pit = m_vpa.find(std::make_pair(v3,v1));

    unsigned v31;
    if (pit != m_vpa.end()) {
      v31 = pit->second;
    }
    else {
      t3DVector pos = eval(c31.m_x,c31.m_y,c31.m_z,f->get_hedge());
      lv1.push_back(pos.m_x);
      lv1.push_back(pos.m_y);
      lv1.push_back(pos.m_z);

      pos = eval_pnt(c31.m_x,c31.m_y,c31.m_z,f);
      lv2.push_back(pos.m_x);
      lv2.push_back(pos.m_y);
      lv2.push_back(pos.m_z);

      v31 = vcount;
      m_vpa.insert(std::make_pair(std::make_pair(v1,v3),v31));
      ++vcount;
    }

    if (lod == 0) {
      lf.push_back(v1);
      lf.push_back(v12);
      lf.push_back(v31);

      lf.push_back(v12);
      lf.push_back(v2);
      lf.push_back(v23);

      lf.push_back(v31);
      lf.push_back(v23);
      lf.push_back(v3);

      lf.push_back(v12);
      lf.push_back(v23);
      lf.push_back(v31);
    }
    else {
      midpointsub(lod-1,vcount,f, v1,v12,v31, c1,c12,c31,lv1,lv2,lf);
      midpointsub(lod-1,vcount,f,v12, v2,v23,c12, c2,c23,lv1,lv2,lf);
      midpointsub(lod-1,vcount,f,v31,v23, v3,c31,c23, c3,lv1,lv2,lf);
      midpointsub(lod-1,vcount,f,v12,v23,v31,c12,c23,c31,lv1,lv2,lf);
    }
  }


}
