/** 
 * @class pps::tPPS
 *  
 * @brief A class implementing a parametric pseudo-surface.
 *
 * @author
 * Marcelo Ferreira Siqueira @n
 * Universidade Federal de Mato Grosso do Sul, @n
 * Departamento de Computacao e Estatistica, @n
 * marcelo at dct (dot) ufms (dot) br
 *
 * @version 1.0
 * @date May 2008
 *
 * @attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */


#ifndef PPS_H_
#define PPS_H_

#include "exceptionobj.h"
#include "smartptr.h"
#include "dataobj.h"

#include "vertex.h"
#include "halfedge.h"
#include "edge.h"
#include "face.h"
#include "complex.h"

#include "3dvector.h"

#include "myheattrib.h"
#include "myvtattrib.h"
#include "myfcattrib.h"

#include <vector>
#include <map>
#include <cmath>
#include <string>
#include <stdlib.h>

namespace pps {

  using dcel::tVertex;
  using dcel::tEdge;
  using dcel::tFace;
  using dcel::tComplex;

  using common::t3DVector;

  /**
   * Class tPPS: A class implementing a parametric pseudo-surface.
   *
   */ 

  class tPPS {
  public:
    /** 
     * Defines vtPointer as a smart pointer to tVertex.
     */
    typedef tVertex::Pointer                 vtPointer;


    /** 
     * Defines hePointer as a smart pointer to tHalfEdge.
     */
    typedef dcel::tHalfEdge::Pointer         hePointer;


    /** 
     * Defines hePointer as a smart pointer to tFace.
     */
    typedef tFace::Pointer                   fcPointer;


    /** 
     * Defines cpPointer as a smart pointer to tComplex.
     */
    typedef tComplex::Pointer                cpPointer;


    /** 
     * Defines vaPointer as a smart pointer to tMyVTAttrib.
     */
    typedef tMyVTAttrib::Pointer             vaPointer;


    /** 
     * Defines haPointer as a smart pointer to tMyHEAttrib.
     */
    typedef tMyHEAttrib::Pointer             haPointer;


    /** 
     * Defines faPointer as a smart pointer to tMyFCAttrib.
     */
    typedef tMyFCAttrib::Pointer             faPointer;


    /** 
     * Defines vtIterator as a iterator of vertices.
     */    
    typedef tComplex::vtIterator             vtIterator;


    /** 
     * Defines fcIterator as a iterator of faces.
     */    
    typedef tComplex::fcIterator             fcIterator;


    /**
     * Defines types for visualization purposes.
     */
    typedef std::map< vtPointer, unsigned >  VNTTAB;
    typedef std::map< std::pair< unsigned, unsigned>, unsigned > VPATAB;


    /**
     * Constructor
     *
     * @param nv: number of vertices of the simplicial surface.
     * @param vset: rectangular  coordinates of the simplicial surface
     * vertices.
     * @param nf: number of faces of the simplicial surface.
     * @param fset: rectangular  coordinates of the simplicial surface
     * faces.
     * @param deg: parametrization degree.
     *
     */
    tPPS(
         unsigned   nv,
         double*    vset,
         unsigned   nf,  
         unsigned*  fset
	)
      throw (tExceptionObject)
    {
      /** 
       * Create a simplicial complex.
       */
      tComplex::Pointer cp = tComplex::NEW(tComplex::atPointer(0));

      try {
        cp->build(nv,vset,nf,fset,3);
      }
      catch (const tExceptionObject& xpt) {
        treatException(xpt);
        exit(0);
      }

      this->set_surface( cp );

      this->set_max_approx_error( 0 );
      this->set_min_approx_error( 0 );
      this->set_rms_approx_error( 0 ); 

      build();
    }


    /** 
     * Destructor.
     */
    ~tPPS() { m_surf->cleanup(); }


    /**
     * Returns  the  maximum error  of  the  approximation  of the  PN
     * triangle surface by this parametric pseudo-surface.
     *
     * @return  The  maximum error  of  the  approximation  of the  PN
     * triangle surface by this parametric pseudo-surface.
     *
     */
    double get_max_approx_error() const
      { return m_maxerr; }


    /**
     * Returns  the  minimum error  of  the  approximation  of the  PN
     * triangle surface by this parametric pseudo-surface.
     *
     * @return  The  minimum error  of  the  approximation  of the  PN
     * triangle surface by this parametric pseudo-surface.
     *
     */
    double get_min_approx_error() const
      { return m_minerr; }


    /**
     * Returns the absolute  RMS error of the approximation  of the PN
     * triangle surface by this parametric pseudo-surface.
     *
     * @return The absolute  RMS error of the approximation  of the PN
     * triangle surface by this parametric pseudo-surface.
     *
     */
    double get_rms_approx_error() const
      { return m_rmserr; }


    /**
     * Returns  a pointer  to the  simplicial surface  associated with
     * this object.
     *
     * @return A pointer to a simplicial surface.
     *
     */
    cpPointer surf() const
      { return m_surf; }


    /**
     * Returns the iterator of  faces of simplicial surface associated
     * with this parametric pseudo-surface. The iterator is set to its
     * first position.
     *
     * @return A face iterator pointing to the first element of a list
     * of faces.
     *
     */
    fcIterator fc_begin() const { return surf()->fc_begin();  }


    /**
     * Returns the iterator of  faces of simplicial surface associated
     * with this parametric pseudo-surface. The iterator is set to its
     * last position.
     *
     * @return A face iterator pointing  to the last element of a list
     * of faces.
     *
     */
    fcIterator fc_end() const { return surf()->fc_end();  }


    /**
     * Method for computing the approximation error.
     *
     */
    void comp_approx_error() ;
 

    /**
     * General  parametric  function  defining this  pseudo-parametric
     * surface at  a point given  by its barycentric  coordinates with
     * respect   to   the  affine   frame   (triangle)  (0,0),   (1/2,
     * -SQRT(3)/2), (1/2, SQRT(3)/2).
     */
    void eval(double, double, double, const fcPointer&, double[3])
      const throw (tExceptionObject);


   /**
    * Samples this  PPS and creates  a "surface mesh" with  the sample
    * points.   This method  is  meant to  be  used for  visualization
    * purposes only.
    *
    */
    void sample(unsigned, std::vector<double>&, std::vector<double>&,
      std::vector<unsigned>&);


  private:
    /**
     * Assigns  a real  number to  the attribute  that  represents the
     * absolute maximum error of  the approximation of the PN triangle
     * surface by this parametric pseudo-surface.
     *
     * @param e: a real number representing the absolute maximum error
     * of  the  approximation  of  the  PN triangle  surface  by  this
     * parametric pseudo-surface.
     *
     */
    void set_max_approx_error( double e )
    { 
      m_maxerr = e; 
    }


    /**
     * Assigns  a real  number to  the attribute  that  represents the
     * absolute minimum error  of the  approximation of  the  PN triangle
     * surface by this parametric pseudo-surface.
     *
     * @param e: a real number representing the absolute minimum error
     * of  the  approximation  of  the  PN triangle  surface  by  this
     * parametric pseudo-surface.
     *
     */
    void set_min_approx_error( double e )
    { 
      m_minerr = e; 
    }


    /**
     * Assigns  a real  number to  the attribute  that  represents the
     * maximum error  of the approximation of the  PN triangle surface
     * by this parametric pseudo-surface.
     *
     * @param e: a  real number representing the maximum  error of the
     * approximation  of the  PN triangle  surface by  this parametric
     * pseudo-surface.
     *
     */
    void set_rms_approx_error( double e )
    { 
      m_rmserr = e; 
    }


    /**
     * Assigns  a  given  pointer   to  the  pointer  to  the  surface
     * associated with this object.
     *
     * @param cp: a pointer to a simplicial surface.
     *
     */
    void set_surface(const cpPointer& cp)
    { 
      m_surf = cp; 
    }


    /** 
     * Builds a parametric pseudo-surface.
     *
     */
    void build() throw (tExceptionObject);


    /**
     * Method for creating half-edge attributes.
     *
     */
    void create_hedge_attribs() const;


    /**
     * Method for creating face attributes.
     *
     */
    void create_face_attribs() const;


    /**
     * Method for creating vertex attributes.
     */
    void create_vertex_attribs() const;


    /**
     * Method for creating a PN triangle surface.
     *
     */
    void create_pnt_surface() const;


    /**
     * Method for creating the P-domain parametrizations.
     *
     */
    void create_param() const;


    /**
     * Method for computing the unit normal of the supporting plane of
     * a given face.
     */
    t3DVector comp_face_normal(const fcPointer&) const;


    /**
     * Method for  sampling a given  P-domain and all the  PN triangle
     * surface patches  associated with the triangles  of the P-domain
     * standard triangulation.
     *
     */
    void build_patch( std::vector<double>&, std::vector<double>&,
      const vtPointer& ) const throw (tExceptionObject);


    /**
     * Method for computing the bi-degree of the parametrizations.
     */
    unsigned compute_patch_degree( unsigned nu ) const
    { return ( ( nu + 1 ) > 7 ) ? 7 : ( nu + 1 ) ; }


    /**
     * Method  for computing  the number  of points  of  the parameter
     * domain used  by the least squares fitting  procedure to compute
     * the parametrizations.
     *
     * @param   d:   (d,d)   is   the   desired   bi-degree   of   the
     * parametrizations
     */
    unsigned compute_number_of_param_points( unsigned d ) const
    { return d << 1 ; }


    /**
     * Method  for sampling a given  P-domain.
     */
    void sample_param_space( const vtPointer&, std::vector<double>& ) 
      const;

    /**
     * Method for sampling the PN triangle surface with parameter points
     * defined in a given P-domain.
     */
    void sample_pnt_surface( const vtPointer&, std::vector<double>&,
      std::vector<double>& ) const throw (tExceptionObject);


    /**
     * Method for sampling the PPS  with parameter points defined in a
     * given P-domain.
     */
    void sample_pps( const vtPointer&, std::vector<double>&,
      std::vector<double>& ) const throw (tExceptionObject);


    /**
     * Method for  finding the triangle of  the standard triangulation
     * of a P-domain  that contains a given point (if  there is such a
     * triangle).
     *
     */
    bool find_triangle( double, double, hePointer&, double&, double&,
      double& ) const;


    /**
     * Auxiliary   method  implementing   a  function   defining  this
     * pseudo-parametric surface  at a point given  by its barycentric
     * coordinates with respect to  the affine frame (triangle) (0,0),
     * (cos(pi/6), -sin(pi/6)), and (cos(pi/6), sin(pi/6)).
     */
    t3DVector eval(double, double, double, const hePointer&) const 
      throw (tExceptionObject);


    /**
     * General parametric function defining a PN triangle surface at a
     * point given by its  barycentric coordinates with respect to the
     * affine  frame (triangle)  (0,0),  (cos(pi/6), -sin(pi/6)),  and
     * (cos(pi/6), sin(pi/6)).
     */
    t3DVector eval_pnt(double, double, double, const fcPointer&)
      const;


    /**
     * Computes  the contribution  of the  surface patch  associated a
     * given P-domain  at a given  point.  The P-domain  is implicitly
     * identified  by a given  half-edge (whose  origin vertex  is the
     * vertex  associated  with  the  P-domain). The  given  point  is
     * expected  to be  in the  P-triangle on  the left  of  the given
     * half-edge.
     */
    void comp_lpd_contrib(const hePointer&, double, double, double&,
      t3DVector&) const throw (tExceptionObject);


    /**
     * Computes  the contribution  of the  surface patch  associated a
     * given P-domain  at a given  point.  The P-domain  is implicitly
     * identified  by a given  half-edge (whose  origin vertex  is the
     * vertex  associated  with  the  P-domain). The  given  point  is
     * expected  to be in  the P-triangle  on the  right of  the given
     * half-edge.
     */
    void comp_rpd_contrib(const hePointer&, double, double, double&,
      t3DVector&) const throw (tExceptionObject);


    /**
     * Converts  the  barycentric  coordinates   of  a  point  in  the
     * reference    triangle   to   the    corresponding   rectangular
     * coordinates.
     */
    void from_bary_to_rect(double, double, double, double, double, 
      double, double, double, double, double&, double&) const;


    /**
     * Computes  the  barycentric coordinates  of  a  given point  (in
     * rectangular  coordinates  with  respect  to a  given  reference
     * triangle.
     */
    void comp_bary_coords(double, double, double, double, double, 
      double, double, double, double&, double&, double&)
      const throw (tExceptionObject);


    /**
     * Evaluates the weight function at a given point.
     */
    double wfunc(double, double, double) const;


    /**
     * Evaluates the  "eta" function at a given  parameter value
     */
    double nfunc(double, double, double) const;


    /**
     * Samples a "curved face" of this surface.
     */
    void midpointsub(unsigned, unsigned&, const fcPointer&, 
      std::vector<double>&, std::vector<double>&, 
      std::vector<unsigned>&);


    /**
     * Samples a "curved face" of this surface.
     */
    void midpointsub(unsigned, unsigned&, const fcPointer&, unsigned,
      unsigned, unsigned, const t3DVector&, const t3DVector&, 
      const t3DVector&, std::vector<double>&,  std::vector<double>&,
      std::vector<unsigned>&);


    /**
     * A pointer to a simplicial surface.
     */
    cpPointer m_surf;


    /**
     * A real  number representing the  maximum absolute error  in the
     * approximation  of the  PN  triangle surface  by the  parametric
     * pseudo-surface.
     */
    double m_maxerr;


    /**
     * A real  number representing the  minimum absolute error  in the
     * approximation  of the  PN  triangle surface  by the  parametric
     * pseudo-surface.
     */
    double m_minerr;


    /**
     * A  real  number  representing  the  RMS of  the  error  in  the
     * approximation  of the  PN  triangle surface  by the  parametric
     * pseudo-surface.
     */
    double m_rmserr;


    /** 
     * The following "maps" are used for visualization purpose only.
     */
    VNTTAB m_vnt;
    VPATAB m_vpa;

  };


}


#endif     // PPS_H_
