//  ===================================================================
//
//  Program:  testpps-sphloop
//  Module:   main.cpp
//  Language: C++
//  Date:     November 2008
//  Version:  1.0.0
//
//  Copyright (c) 2008
//
//  Marcelo Siqueira
//
//  Departamento de Computao e Estatstica
//  Universidade Federal de Mato Grosso do Sul
//  Av. Costa e Silva, S/N, Campus Universitria
//  Campo Grande, MS, Brasil, 79090-700
//  Phone No.: +55 67 3345 7515
//
//  marcelo@dct.ufms.br
//
// This program  is distributed  WITHOUT ANY WARRANTY,  and it  may be
// freely redistributed under the condition that the copyright notices
// are  not  removed,  and  no  compensation  is  received.   Private,
// research, and institutional use  is free. Distribution of this code
// as  part of  a  commercial  system is  permissible  ONLY BY  DIRECT
// ARRANGEMENT WITH THE AUTHOR.
//
//  ===================================================================

#include <iostream>
#include <sstream>
#include <string>
#include <stdlib.h>
#include <vector>
#include <sys/time.h>

#include "exceptionobj.h"
#include "macros.h"

#include "reader.h"
#include "smartptr.h"
#include "dataobj.h"
#include "complex.h"
#include "cpattrib.h"
#include "pps.h"
#include "writer.h"


int main(int argc, char *argv[])
{
  using std::cout;
  using std::endl;

  using offlib::tReader;
  using dcel::tComplex;
  using pps::tPPS;
  using offlib::tWriter;

  using pps::tMyHEAttrib;

  if (argc != 4) {
    cout << endl;
    cout << "Usage: testpps-sphloop <arg 1> <arg 2> <arg 3>" << endl;
    cout << endl;
    cout << "arg1: filename of an OFF file without extension." << endl;
    cout << "arg2: filename of the Loop subdivision eigenstructure data table ." << endl;
    cout << "arg4: the level-of-detail of the output triangular mesh." << endl;
    cout << endl;
    cout << "For instance:" << endl;
    cout << "testpps-sphloop examples/retinal lpdata50.dat 3" << endl;
    cout << endl;
    return 0;
  }

  std::string fn1(argv[1]);
  std::string fn4(argv[2]);
  unsigned lod = unsigned(atoi(argv[3]));

  cout << "Reading input file..." << endl;

  struct timeval stime, etime;

  gettimeofday(&stime,0);

  unsigned nv, nf;
  double* vset;
  unsigned* fset;

  try {
    tReader reader( fn1 );
    reader.read( nv , vset , nf , fset );
  }
  catch (const tExceptionObject& xpt) {
    treatException(xpt);
    exit(0);
  }

  gettimeofday(&etime,0);

  long tt = 1000l * (etime.tv_sec - stime.tv_sec) 
          + (etime.tv_usec - stime.tv_usec) / 1000l;

  cout << "Input file read in " 
       << tt 
       << " miliseconds." 
       << endl;

  cout << endl;

  cout << "Building the parametric pseudo-surface..." << endl;

  gettimeofday(&stime,0);

  tPPS* pps = 0;

  try {
    pps = new tPPS( nv , vset , nf , fset , fn4 );
  }
  catch (const tExceptionObject& xpt) {
    treatException(xpt);
    exit(0);
  }

  gettimeofday(&etime,0);

  tt = 1000l * (etime.tv_sec - stime.tv_sec) 
     + (etime.tv_usec - stime.tv_usec) / 1000l;

  cout << "Smooth surface created in " 
       << tt 
       << " miliseconds." 
       << endl;

  cout << endl;

  cout << "Releasing memory..." << endl;

  if (vset != 0) {
    delete[] vset;
  }

  if (fset != 0) {
    delete[] fset;
  }

  cout << endl;

  cout << "Computing the approximation error of the least squares fitting..." 
       << endl;

  gettimeofday(&stime,0);

  pps->comp_approx_error();

  gettimeofday(&etime,0);

  tt = 1000l * (etime.tv_sec - stime.tv_sec) 
     + (etime.tv_usec - stime.tv_usec) / 1000l;

  cout << "Maximum error: " << pps->get_max_approx_error() << endl;

  cout << "Minimum error: " << pps->get_min_approx_error() << endl;

  cout << "Root mean square (RMS): " << pps->get_rms_approx_error() << endl;

  cout << "Error measures computed in " 
       << tt 
       << " miliseconds." 
       << endl;

  cout << endl;

  cout << "Generating output files..." << endl;

  gettimeofday(&stime,0);

  std::vector<double> lv1, lv2;
  std::vector<unsigned> lf;

  pps->sample( lod , lv1 , lv2 , lf );

  std::string fn2( fn1 + std::string("-sphloop-pps-") + std::string(argv[3]) );

  std::string fn3( fn1 + std::string("-sphloop-") + std::string(argv[3]) );

  try {
    nv = unsigned(lv1.size()) / 3;
    nf = unsigned( lf.size()) / 3;

    tWriter writer1(fn2);
    writer1.write(nv, (const double*) &lv1[0], nf, (const unsigned*) &lf[0]);

    lv1.clear();

    tWriter writer2(fn3);
    writer2.write(nv, (const double*) &lv2[0], nf, (const unsigned*) &lf[0]);

    lv2.clear();
    lf.clear();
  }
  catch (const tExceptionObject& xpt) {
    treatException(xpt);
    exit(0);
  }

  gettimeofday(&etime,0);

  tt = 1000l * (etime.tv_sec - stime.tv_sec) 
     + (etime.tv_usec - stime.tv_usec) / 1000l;

  cout << "Output files generated in " 
       << tt 
       << " miliseconds." 
       << endl;

  cout << endl;

  cout << "Releasing memory..." << endl;

  if (pps != 0) {
    delete pps;
  }

  cout << endl;

  cout << "Done!" << endl;

  cout << endl;

  return 0;
}
